{***********************************************************}
{   This file is part of the DORB project.                  }
{                                                           }
{   See the file License, included in this distribution,    }
{   for details about the copyright.                        }
{                                                           }
{   Send comments and/or bug reports to: vic@inec.ru        }
{***********************************************************}
unit codeunit;

interface

uses
  codebase,classes,sysutils,ir_int,code_int,orbtypes,
  codeutil,{$IFDEF MSWINDOWS}windows,comobj,{$ENDIF}ir,idldb;

type

  TCodegen = class(TCodebase)
  private
    FConstants : IInterfaceList;
  protected
    // interface unit
    procedure int_declare_alias(cont : IAliasDef);
    procedure int_declare_array(arr: IArrayDef);
    procedure int_declare_sequence(seq: ISequenceDef);
    procedure int_declare_enum(enum : IEnumDef);
    procedure int_declare_interface(int: IInterfaceDef);
    procedure int_declare_struct(str: IStructDef);
    procedure int_declare_value(val: IValueDef);
    procedure int_declare_module(m: IModuleDef);
    procedure int_declare_union(u: IUnionDef);
    procedure int_declare_exception(e : IExceptionDef);
    procedure int_declare_internal_type(cont_name,name: string; t: IIDLType);

    // code unit declarations
    procedure code_declare_alias(cont : IAliasDef);
    procedure code_declare_enum(enum : IEnumDef);
    procedure code_declare_interface(int: IInterfaceDef);
    procedure code_declare_struct(str: IStructDef);
    procedure code_declare_value(val: IValueDef);
    procedure code_declare_module(m: IModuleDef);
    procedure code_declare_union(u: IUnionDef);
    procedure code_declare_array(a: IArrayDef);
    procedure code_declare_sequence(s: ISequenceDef);
    procedure code_declare_exception(e: IExceptionDef);


    // code unit implementations
    procedure code_impl_alias(cont : IAliasDef);
    procedure code_impl_enum(enum : IEnumDef);
    procedure code_impl_struct(str: IStructDef);
    procedure code_impl_interface(int: IInterfaceDef);
    procedure code_impl_value(val: IValueDef);
    procedure code_impl_module(m: IModuleDef);
    procedure code_impl_union(u: IUnionDef);
    procedure code_impl_exception(e: IExceptionDef);

    // dependencies
    procedure dependencies(contents: IInterfaceList);
    procedure int_dependencies(int: IInterfaceDef);
    procedure union_dependencies(u: IUnionDef);
    procedure alias_dependencies(val: IAliasDef);
    procedure array_dependencies(a: IArrayDef);
    procedure seq_dependencies(s: ISequenceDef);
    procedure op_dependencies(oper: IOperationdef);
    procedure attr_dependencies(attr: IAttributeDef);
    procedure val_dependencies(val: IValueDef);
    procedure cont_dependencies(val: IIRObject);
    procedure struct_dependencies(val: IStructDef);

    procedure collect_constant(con: IConstantDef);
    procedure declare_constant(con: IConstantDef);
    procedure declare_forwards();
    procedure ex_dependencies(ex: IExceptionDef);

    // impl unit declarations
    procedure final_declare_interface(int: IInterfaceDef);
    procedure final_declare_value(val: IValueDef);

    // impl unit implementations
    procedure final_impl_interface(int: IInterfaceDef);
    procedure final_impl_value(val: IValueDef);

    procedure impl_code(cont: IIRObject);
    procedure impl_declare(cont: IIRObject);
    procedure int_declare(cont: IIRObject);
    procedure write_header();
    function get_default_value_for_union(def: IIDLType;members: UnionMemberSeq): Variant;

    procedure final_code(cont: IContained);
    procedure final_declare(cont: IContained);
  public
    constructor Create(rep: IRepository; params: TStrings;db: TDB);
    destructor Destroy(); override;
    procedure generate;
  end;

implementation

uses
  orb_int;

//*********************************************************************
// TCodegen
//*********************************************************************
constructor TCodegen.Create(rep: IRepository; params: TStrings; db: TDB);
begin
  FRepository := rep;
  FParams := params;
  FDB := db;
  FDeclared := TStringList.Create;
  FForwards := TInterfaceList.Create;
  Marshallers := TStringList.Create;
  FImplemented := TStringList.Create;
  FConstants := TInterfaceList.Create;
  FIncludes := TStringList.Create;
end;

destructor TCodegen.Destroy();
begin
  FDeclared.free;
  FImplemented.free;
  Marshallers.free;
  FIncludes.free;
end;

procedure TCodegen.generate;
var
  temp,name,usesstr: string;
  i,cnt : integer;
  contents: IInterfaceList;
  it : IIDLType;

  function _ExtractFileName(str: string): string;
  var
    I: Integer;
  begin
    I := LastDelimiter('\:/', str);
    Result := Copy(str, I + 1, MaxInt);
  end;

begin
  try
    name := _ExtractFileName(FParams[0]);
    if Pos('.',name) > 0 then
      name := Copy(name,1,Pos('.',name)-1);
    contents := FRepository.contents();
    // interfaces
    assign(output,name+'_int.pas');
    Rewrite(output);
    write_header();
    //writeln('unit '+name+'_int;'+char(CR));
    writeln('unit '+name+'_int;'); writeln;
    //writeln('{$MINENUMSIZE 4}'+char(CR));
    writeln('{$MINENUMSIZE 4}'); writeln;
    //writeln('interface'+ char(CR));
    writeln('interface'); writeln;
    writeln('uses');
    dependencies(contents);
    FDeclared.Clear;
    usesstr := '  code_int, orb_int, orbtypes';
    if FIncludes.count > 0 then
      begin
        cnt := FIncludes.count-1;
        for i := 0 to cnt do
          begin
            if (length(usesstr) > 80) and (i <> cnt) then
              begin
                writeln(usesstr+',');
                usesstr := '  '+FIncludes[i] + '_int';
              end
            else
              usesstr := usesstr + ', ' + FIncludes[i] + '_int';
            if i = cnt then usesstr := usesstr+';';
          end
      end
    else
      usesstr := usesstr + ';';
    writeln(usesstr);
    writeln;

    if FConstants.Count > 0 then
      begin
        i := 0;
        while i < FConstants.count do begin
          if not FDB.repoid_defined((FConstants[i] as IConstantDef).id) then Break;
          i := Succ(i);
        end; { while }
        if i < FConstants.count then
          //writeln('const'+ char(CR));
          writeln('const'); writeln;
        for i := 0 to Pred(FConstants.count) do
          declare_constant(FConstants[i] as IConstantDef);
        writeln;
      end;

    //writeln('type'+ char(CR));
    writeln('type'); writeln;

    declare_forwards();

    for i := 0 to Pred(contents.count) do
      int_declare(IContained(contents[i]));

    //writeln('implementation'+ char(CR));
    writeln('implementation'); writeln;
    writeln('end.');

    Close(output);
    assign(output,name+'.pas');
    Rewrite(output);

    // implementations
    write_header();
    //writeln('unit '+name+';'+char(CR));
    writeln('unit '+name+';'); writeln;
    //writeln('interface'+ char(CR));
    writeln('interface'); writeln;
    if FIncludes.IndexOf('except') >= 0 then FIncludes.Delete(FIncludes.IndexOf('except'));
    if FParams.IndexOf('-poa') >= 0 then FIncludes.add('poa');
    writeln('uses');
    writeln('  Classes, orb_int, orb, req_int, code_int, imr, imr_int, env_int, stdstat, std_seq,');
    usesstr :='  orbtypes, exceptions, except_int, '+name+'_int';
    if FIncludes.count > 0 then
      begin
        cnt := FIncludes.count-1;
        for i := 0 to cnt do
          begin
            if (length(usesstr) > 80) then
              begin
                writeln(usesstr+',');
                usesstr := '  '+FIncludes[i] + '_int';
              end
            else
              usesstr := usesstr + ', ' + FIncludes[i] + '_int';
            if (length(usesstr) > 80) and (i <> cnt) then
              begin
                writeln(usesstr+',');
                usesstr := '  '+FIncludes[i];
              end
            else
              usesstr := usesstr + ', ' + FIncludes[i];
            if i = cnt then usesstr := usesstr+';';
          end
      end
    else
      usesstr := usesstr + ';';
    writeln(usesstr);
    writeln;
    //writeln('type'+ char(CR));
    writeln('type'); writeln;

    for i := 0 to Pred(contents.count) do
      impl_declare(contents[i] as IIRObject);

    if Marshallers.count > 0 then
      begin
        if FParams.IndexOf('-any') > 0 then
          for i := 0 to Pred(Marshallers.count) do
            begin
              if IIRObject(Pointer(Marshallers.Objects[i])).def_kind in [dk_Array,dk_Sequence] then
                temp := ''
              else
                temp := ': '+type_str(IIRObject(Pointer(Marshallers.Objects[i])));
              writeln('  function '+Marshallers[i]+'_to_any(const val'+temp+'): IAny;');
              writeln('  function any_to_'+Marshallers[i]+'(const a: IAny; var val'+temp+'): Boolean;');
              writeln;
            end;
        writeln('var');
        for i := 0 to Pred(Marshallers.count) do
          writeln('  '+ Marshallers[i]+ '_marshaller : IStaticTypeInfo;');
        writeln;
      end;

    //writeln('implementation'+ char(CR));
    writeln('implementation'); writeln;
    writeln('uses');
    //writeln('  any,tcode,static,SysUtils,throw;'+ char(CR));
    writeln('  any,tcode,static,SysUtils,throw;'); writeln;

    if (Marshallers.count > 0) and (FParams.IndexOf('-any') > 0) then
        begin
          writeln('var');
          for i := 0 to Pred(Marshallers.count) do
            writeln('  _tc_'+Marshallers[i]+' : ITypeCodeConst;');
          writeln;
        end;


    // implementation
    for i := 0 to Pred(contents.count) do
      impl_code(contents[i] as IIRObject);

    if Marshallers.count > 0 then
      begin
        if FParams.IndexOf('-any') > 0 then
          begin
            for i := 0 to Pred(Marshallers.count) do
              begin
                if IIRObject(Pointer(Marshallers.Objects[i])).def_kind in [dk_Array,dk_Sequence] then
                  temp := ''
                else
                  temp := ': '+type_str(IIRObject(Pointer(Marshallers.Objects[i])));
                writeln('function any_to_'+Marshallers[i]+'(const a: IAny; var val'+temp+'): Boolean;');
                writeln('var');
                writeln('  stat: IStaticAny;');
                writeln('begin');
                writeln('  stat := StaticAny('+Marshallers[i]+'_marshaller,@val);');
                writeln('  result := a.to_static_any(stat);');
                writeln('end;');
                writeln;
                writeln('function '+Marshallers[i]+'_to_any(const val'+temp+'): IAny;');
                writeln('var');
                writeln('  stat: IStaticAny;');
                writeln('begin');
                writeln('  stat := StaticAny('+Marshallers[i]+'_marshaller,@val);');
                writeln('  result := CreateAny;');
                writeln('  result.from_static_any(stat);');
                writeln('end;');
                writeln;
              end;
          end;
        writeln('initialization');
        for i := 0 to Pred(Marshallers.count) do
          writeln('  '+ Marshallers[i]+ '_marshaller := T'+ Marshallers[i]+ '_marshaller.Create();');
        if FParams.IndexOf('-any') > 0 then
          for i := 0 to Pred(Marshallers.count) do
            begin
              it := TIDLType._narrow(IIRObject(Pointer(Marshallers.Objects[i])) as IORBObject);
              if it = nil then
                assert(false,'not implemented')
              else
                begin
                  write('  _tc_'+ Marshallers[i] + ' := CreateTypeCodeConst(');
                  temp := it._type.to_string;
                  while length(temp) > 120 do
                    begin
                      writeln(''''+copy(temp,1,110)+'''+');
                      write('    ');
                      temp := copy(temp,111,10000);
                    end;
                  writeln(''''+temp+''');');
                end;
            end;
      end;

    writeln('end.');
    Close(output);
    if FParams.IndexOf('-impl') >= 0 then
      begin
        assign(output,name+'_impl.pas');
        Rewrite(output);
        write_header();
        //writeln('unit '+name+'_impl;'+char(CR));
        writeln('unit '+name+'_impl;'); writeln;
        //writeln('interface'+ char(CR));
        writeln('interface'); writeln;
        writeln('uses');
        writeln('  orb_int, orb, req_int, code_int, imr, imr_int, env_int, except_int, orbtypes,');
        usesstr :='  exceptions, stdstat, std_seq, Classes, '+name+'_int, '+name;
        if FIncludes.count > 0 then
          begin
            cnt := FIncludes.count-1;
            for i := 0 to cnt do
              begin
                if (length(usesstr) > 80) and (i <> cnt) then
                  begin
                    writeln(usesstr+',');
                    usesstr := '  '+FIncludes[i]+'_int';
                  end
                else
                  usesstr := usesstr + ', ' + FIncludes[i]+'_int';
                if i = cnt then usesstr := usesstr+';';
              end
          end
        else
          usesstr := usesstr + ';';
        writeln(usesstr);
        writeln;
        //writeln('type'+ char(CR));
        writeln('type'); writeln;

        for i := 0 to Pred(contents.count) do
          final_declare(IContained(contents[i]));

        writeln('implementation');
        writeln;

        for i := 0 to Pred(contents.count) do
          final_code(IContained(contents[i]));

        writeln('end.');
        Close(output);
      end;
    Assign(output,'');
  except
    on e : Exception do
      begin
        Close(output);
        Assign(output,'');
        write('error: '+e.Message);
      end;
  end;
end;

{** calls interface declaration method for each
    definition kind }
procedure TCodegen.int_declare(cont: IIRObject);
begin
  case cont.def_kind of
    dk_Interface : int_declare_interface(cont as IInterfaceDef);
    dk_Struct: int_declare_struct(cont as IStructDef);
    dk_Alias : int_declare_alias(cont as IAliasDef);
    dk_Union : int_declare_union(cont as IUnionDef);
    dk_Enum : int_declare_enum(cont as IEnumDef);
    dk_Value : int_declare_value(cont as IValueDef);
    dk_Module : int_declare_module(cont as IModuleDef);
    dk_Exception : int_declare_exception(cont as IExceptionDef);
  end;
end;

{** calls class declaration method for each
    definition kind }
procedure TCodegen.impl_declare(cont: IIRObject);
//var
  //t : IIDLType;
  //c: IContained;
begin
  case cont.def_kind of
    dk_Interface : code_declare_interface(cont as IInterfaceDef);
    dk_Struct: code_declare_struct(cont as IStructDef);
    dk_Enum: code_declare_enum(cont as IEnumDef);
    dk_Alias : impl_declare((cont as IAliasDef).original_type_def);
    dk_Value : code_declare_value(cont as IValueDef);
    dk_Module : code_declare_module(cont as IModuleDef);
    dk_Union : code_declare_union(cont as IUnionDef);
    dk_Exception : code_declare_exception(cont as IExceptionDef);
    dk_Array : code_declare_array(cont as IArrayDef);
    dk_Sequence : code_declare_sequence(cont as ISequenceDef);
  end;
end;

{** calls class implementation method for each
    definition kind }
procedure TCodegen.impl_code(cont: IIRObject);
begin
  case cont.def_kind of
    dk_Interface : code_impl_interface(cont as IInterfaceDef);
    dk_Struct : code_impl_struct(cont as IStructDef);
    dk_Enum : code_impl_enum(cont as IEnumDef);
    dk_Alias: code_impl_alias(cont as IAliasDef);
    dk_Value : code_impl_value(cont as IValueDef);
    dk_Module : code_impl_module(cont as IModuleDef);
    dk_Union : code_impl_union(cont as IUnionDef);
    dk_Exception : code_impl_exception(cont as IExceptionDef);
  end;
end;

{** prints compiler information on top of each generated unit }
procedure TCodegen.write_header();
begin
  //writeln('// Generated by DORB idltopas compiler, '+FormatDateTime('dd/mm/yyyy',now));
  writeln('// **********************************************************************');
  writeln('//                                                                       ');
  writeln('// Generated by the MTDORB IDL-to-Pascal Translator, '+FormatDateTime('dd/mm/yyyy hh:nn',now));
  writeln('//                                                                       ');
  writeln('// Copyright (c) 2001                                                    ');
  writeln('// Millennium Group.                                                     ');
  writeln('// Samara, Russia                                                        ');
  writeln('//                                                                       ');
  writeln('// All Rights Reserved                                                   ');
  writeln('//                                                                       ');
  writeln('// **********************************************************************');
end;

{** checks dependencies for operation definition }
procedure TCodegen.op_dependencies(oper: IOperationdef);
var
  params : ParDescriptionSeq;
  i : integer;
  paramdesc : ParameterDescription;
  cont : IContained;
begin
  if oper._result.kind in [tk_objref,tk_struct,tk_union] then
    begin
      cont := TContained._narrow(oper.get_result_def as IORBObject);
      if (cont <> nil) and (FDeclared.indexOf(cont._absolute_name) < 0) then
        add_forwards(oper.get_result_def as IContained);
    end;
  params := oper.get_params;
  for i := 0 to Pred(Length(params)) do
    begin
      paramdesc := params[i];
      if paramdesc._type.kind in [tk_objref,{tk_struct,}tk_union] then
        begin
          cont := TContained._narrow(paramdesc.type_def as IORBObject);
          if cont = nil then continue;
          if FDeclared.indexOf(cont._absolute_name) < 0 then
            add_forwards(cont);
        end;
    end;
end;

{** checks dependencies for value definition }
procedure TCodegen.val_dependencies(val: IValueDef);
begin
  if FIncludes.IndexOf('value') < 0 then
    FIncludes.add('value');
end;

procedure TCodegen.ex_dependencies(ex: IExceptionDef);
var
  i : integer;
  members : StructMemberSeq;
  cont,defin : IContained;
begin
  members := nil;
  if not declare((ex as IContained)._absolute_name) then exit;
  if FIncludes.IndexOf('except') < 0 then
    FIncludes.add('except');
  members := ex.members^;
  defin := TContained._narrow(ex.defined_in as IORBObject);
  for i := 0 to Pred(Length(members)) do
    begin
      if members[i].type_def.def_kind in [dk_Interface,dk_Value] then
        begin
          cont := TContained._narrow(members[i].type_def as IORBObject);
          if cont = nil then continue;
          if defin = cont then
            begin
              if FDB.repoid_defined(cont.id) then exit;
              if FForwards.IndexOf(cont) < 0 then FForwards.add(cont);
            end
          else if (FDeclared.indexOf(cont._absolute_name) < 0) then
            add_forwards(cont);
        end;
      cont_dependencies(members[i].type_def);
    end;
end;

{** checks dependencies for interface definition }
procedure TCodegen.int_dependencies(int: IInterfaceDef);
var
  i : integer;
  list : IInterfaceList;
begin
  if not declare((int as IContained)._absolute_name) then exit;
  // check exceptions
  list := (int as IContainer).contents(dk_Exception);
  for i := 0 to pred(list.count) do
    ex_dependencies(list[i] as IExceptionDef);
  // check constants
  list := (int as IContainer).contents(dk_Constant);
  for i := 0 to pred(list.count) do
    collect_constant(list[i] as IConstantDef);
  // check forward declarations
  list := int.get_base_interfaces;
  for i := 0 to pred(list.count) do
    int_dependencies(list[i] as IInterfaceDef);
  list := (int as IContainer).contents(dk_Operation);
  for i := 0 to pred(list.count) do
    op_dependencies(list[i] as IOperationdef);
  list := (int as IContainer).contents(dk_Attribute);
  for i := 0 to pred(list.count) do
    attr_dependencies(list[i] as IAttributeDef);
end;

procedure TCodegen.alias_dependencies(val: IAliasDef);
begin
  cont_dependencies(val.original_type_def);
end;

procedure TCodegen.cont_dependencies(val: IIRObject);
begin
  case val.def_kind of
    dk_Interface : int_dependencies(val as IInterfaceDef);
    dk_Value : val_dependencies(val as IValueDef);
    dk_Module : dependencies((val as IContainer).contents);
    dk_Constant : collect_constant(val as IConstantDef);
    dk_Union : union_dependencies(val as IUnionDef);
    dk_Array : array_dependencies(val as IArrayDef);
    dk_Sequence : seq_dependencies(val as ISequenceDef);
    dk_Alias : alias_dependencies(val as IAliasDef);
    dk_Struct : struct_dependencies(val as IStructDef);
    dk_Exception : ex_dependencies(val as IExceptionDef);
  end;
end;

{** calls dependencies ckeck method for each definition kind }
procedure TCodegen.dependencies(contents: IInterfaceList);
var
  i,j : integer;
  id,fname : string;

  function _extractfilename(name: string): string;
  begin
    result := extractfilename(name);
    while pos('/',result) > 0 do
      result := copy(result,pos('/',result)+1,1000)
  end;
begin
  for i := 0 to pred(contents.count) do
    begin
      id := (contents[i] as IContained).id;
      if Fdb.repoid_defined(id) then
        begin
          fname := _extractfilename(Fdb.repoid_file_name(id));
          j := pos('.',fname);
          if j > 1 then
            fname := copy(fname,1,j-1);
          if FIncludes.IndexOf(fname) < 0 then
            FIncludes.add(fname);
        end;
      cont_dependencies(contents[i] as IContained);
    end;
end;

procedure TCodegen.code_declare_alias(cont: IAliasDef);
begin
  if FDB.repoid_defined((cont as IContained).id) then exit;
  declare_marshaller(cont);
end;

procedure TCodegen.code_declare_enum(enum: IEnumDef);
begin
  if FDB.repoid_defined((enum as IContained).id) then exit;
  declare_marshaller(enum);
end;

procedure TCodegen.code_declare_module(m: IModuleDef);
var
  i : integer;
  contents : IInterfaceList;
begin
  //if FDB.repoid_defined(m.id) then exit;
  contents := (m as IContainer).contents();
  for i := 0 to Pred(contents.count) do
    impl_declare(contents[i] as IIRObject);
end;

procedure TCodegen.code_declare_interface(int: IInterfaceDef);
var
  name,iname,cname,basenames,bname,boname: string;
  base: IInterfaceList;
  i : integer;
  stub_name,skel_name , mdirective: string;
  contents : IInterfaceList;

begin
  if FDB.repoid_defined(int.id) then exit;
  contents := (int as IContainer).contents();
  // internal declarations
  for i := 0 to Pred(contents.count) do
    impl_declare(contents[i] as IIRObject);
  name := int._absolute_name;
  iname := 'I'+name;
  cname := 'T'+name;
  stub_name := cname+ '_stub';
  base := int.get_base_interfaces;
  basenames := '';
  boname := 'TORBObject';
  for i := 0 to Pred(base.count) do
    begin
      bname := (base[i] as IContained)._absolute_name;
      if i = 0 then
        boname := 'T'+bname;
      if i <> 0 then
        baseNames := baseNames + ',I'+bname;
    end;
  writeln('  {** '+int.id+' } ');
  writeln('  '+ cname + ' = class('+boname+','+iname+basenames+')');
  writeln('  protected');
  declare_all_methods(int,' virtual; abstract;',true);
  writeln('    function narrow_helper(const str: string): Pointer; override;');
  writeln('  public');
  writeln('    class function _narrow(const obj : IORBObject): '+iname+';');
  writeln('    class function narrow_helper2(const obj: IORBObject): Boolean;');
  writeln('  end;');
  writeln;

  //-----------  stub --------------------
  writeln('  {** stub for interface '+int.id+'} ');
  mdirective := ' override;';
  if base.count = 0 then
    writeln('  '+ stub_name + ' = class('+cname+')')
  else
    begin
      //mdirective := '';
      bname := (base[0] as IContained)._absolute_name;
      System.write('  '+ stub_name + ' = class(');
      System.write(cname+',I'+name);
      for i := 1 to Pred(base.count) do
        begin
          bname := (base[i] as IContained)._absolute_name;
          System.write(',I'+bname);
        end;
      writeln(')');
    end;
  writeln('  protected');
  declare_all_methods(int,mdirective,false);
  writeln('  end;');
  writeln;
  if FParams.IndexOf('-poa') < 0 then
    // BOA
    begin
      writeln('  {** skeleton for interface '+int.id+'} ');
      skel_name := cname+ '_skel';
      //-----------  skeleton --------------------
      if base.count = 0 then
        writeln('  '+ skel_name + ' = class(TStaticMethodDispatcher,IStaticInterfaceDispatcher,'+iname+')')
      else
        begin
          bname := (base[0] as IContained)._absolute_name;
          write('  '+ skel_name + ' = class(T'+bname+'_skel,'+iname);
          for i := 1 to Pred(base.count) do
            begin
              bname := (base[i] as IContained)._absolute_name;
              write(',I'+bname);
            end;
          writeln(')');
        end;
      writeln('  protected');
      declare_all_methods(int,' virtual; abstract;',true);
      if base.count = 0 then
        writeln('    function _dispatch(const req: IStaticServerRequest; const env: IEnvironment): Boolean; virtual;')
      else
        writeln('    function _dispatch(const req: IStaticServerRequest; const env: IEnvironment): Boolean; override;');
      writeln('    function narrow_helper(const str: string): Pointer; override;');
      writeln('  public');
      writeln('    constructor Create();');
      writeln('  end;');
      writeln;
    end
  else
    // POA
    begin
      //-----------  POA stub --------------------
      writeln('  {** POA stub for interface '+int.id+'} ');
      if base.count = 0 then
        writeln('  '+ cname + '_stub_clp = class(TPOAStub,IPOAStub,'+iname+')')
      else
        begin
          bname := (base[0] as IContained)._absolute_name;
          write('  '+ cname + '_stub_clp = class(T'+bname+'_stub_clp,'+iname);
          for i := 1 to Pred(base.count) do
            begin
              bname := (base[i] as IContained)._absolute_name;
              write(',I'+bname);
            end;
          writeln(')');
        end;
      writeln('  protected');
      declare_all_methods(int,' virtual;',True);
      writeln('    function narrow_helper(const str: string): Pointer; override;');
      writeln('  end;');
      writeln;
      //-----------  servant --------------------
      writeln('  {** servant for interface '+int.id+'} ');
      if base.count = 0 then
        writeln('  '+ cname + '_serv = class(TPOAStaticImplementation,IPOAStaticImplementation,IServant,'+iname+')')
      else
        begin
          bname := (base[0] as IContained)._absolute_name;
          write('  '+ cname + '_serv = class(T'+bname+'_serv,'+iname);
          for i := 1 to Pred(base.count) do
            begin
              bname := (base[i] as IContained)._absolute_name;
              write(',I'+bname);
            end;
          writeln(')');
        end;
      if FParams.IndexOf('-notabs') >= 0 then begin
        writeln('  private');
        writeln('    FImplementation: ' + iname + ';');
      end;
      writeln('  protected');
      if FParams.IndexOf('-notabs') >= 0 then
        declare_all_methods(int,' virtual;',true)
      else
        declare_all_methods(int,' virtual; abstract;',true);
      writeln('    function _primary_interface(const objid: ObjectID; const poa: IPOA): string; override;');
      writeln('    function _make_stub(const poa: IPOA; const obj : IORBObject): IORBObject; override;');
      writeln('    function _is_a(const repoid: RepositoryID): Boolean; override;');
      writeln('    procedure invoke(const serv: IStaticServerRequest); override;');
      writeln('    function _dispatch(const req: IStaticServerRequest): Boolean;');
      writeln('    function narrow_helper(const str: string): Pointer; override;');
      writeln('  public');
      if FParams.IndexOf('-notabs') >= 0 then
        writeln('    constructor Create(const AImpl: ' + iname + ' = nil);');
      writeln('    function _this(): '+iname+';');
      writeln('    class function _narrow(const srv : IServant): '+iname+';');
      writeln('  end;');
      writeln;
    end;
  marshaller_decl(name,int)
end;

procedure TCodegen.code_declare_exception(e: IExceptionDef);
var
  name,typestr: string;
  member: StructMember;
  members: StructMemberSeq;
  i,mlen : integer;
begin
  members := nil;
  if FDB.repoid_defined(e.id) then exit;
  name := (e as IContained)._absolute_name;
  marshaller_decl(name,e);
  members := e.members^;
  mlen := Length(members);
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : code_declare_sequence(member.type_def as ISequenceDef);
        dk_Array : code_declare_array(member.type_def as IArrayDef);
      end;
    end;
  writeln('  {** '+(e as IContained).id+' } ');
  writeln('  T'+name+' = class(UserException,I'+name+')');
  if mlen > 0 then  writeln('  private');
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : typestr := 'T'+name + '_'+member.name + '_array';
        else
          typestr := type_str(member.type_def)
      end;
      writeln('    F'+member.name+' : '+ typestr+';');
    end;
  writeln('  protected');
  writeln('    procedure throw; override;');
  writeln('    function clone(): IORBException; override;');
  writeln('    function repoid(): PChar; override;');
  writeln('    procedure encode(const enc: IEncoder); override;');
  ex_members_decl(name,members,ed_in);
  writeln('  public');
  ex_members_decl(name,members,ed_out);
  writeln('    constructor Create(ex: T'+name+'); overload;');
  write('    constructor Create(');
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : continue;
        else
          typestr := type_str(member.type_def)
      end;
      write('_'+member.name+': '+ typestr+' = '+def_value(member.type_def));
      if i < mlen-1 then write(';');
    end;
  writeln('); overload;');
  writeln('  end;');
  writeln;
end;

procedure TCodegen.code_declare_sequence(s: ISequenceDef);
var
  name: string;
  t : IIDLType;
begin
  if FDB.repoid_defined(FDB.pseudo_repoid(s)) then  exit;
  if is_base_sequence(s) then exit;
  name := expand_name(s);
  t := original_type(s.element_type_def);
  if t.def_kind = dk_array then
    code_declare_array(t as IArrayDef);
  marshaller_decl(name,s);
end;

procedure TCodegen.code_declare_array(a: IArrayDef);
var
  name: string;
  t : IIDLType;
begin
  if FDB.repoid_defined(FDB.pseudo_repoid(a)) then  exit;
  name := expand_name(a);
  t := original_type(a.element_type_def);
  if t.def_kind = dk_array then
    code_declare_array(t as IArrayDef);
  marshaller_decl(name,a);
end;

procedure TCodegen.code_declare_union(u: IUnionDef);
var
  name,typestr,last_name: string;
  member: UnionMember;
  members: UnionMemberSeq;
  i : integer;
begin
  members := nil;
  if FDB.repoid_defined((u as IContained).id) then exit;
  name := (u as IContained)._absolute_name;
  members := u.members^;
  for i := 0 to Pred(Length(members)) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : code_declare_sequence(member.type_def as ISequenceDef);
        dk_Array : code_declare_array(member.type_def as IArrayDef);
        dk_Struct : code_declare_struct(member.type_def as IStructDef);
      end;
    end;
  writeln('  {** '+(u as IContained).id+' } ');
  writeln('  T'+name+' = class(TInterfacedObject,I'+name+')');
  writeln('  private');
  last_name := '';
  for i := 0 to Pred(Length(members)) do
    begin
      member := members[i];
      if member.name = last_name then continue;
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : typestr := 'T'+name + '_'+member.name + '_array';
        else
          typestr := type_str(member.type_def)
      end;
      writeln('    F'+member.name+': '+typestr+';');
      last_name := member.name;
    end;
  writeln('    F'+'discriminator: '+type_str(u.discriminator_type_def)+';');
  writeln('  protected');
  last_name := '';
  for i := 0 to Pred(Length(members)) do
    begin
      member := members[i];
      if member.name = last_name then continue;
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : typestr := 'T'+name + '_'+member.name + '_array';
        else
          typestr := type_str(member.type_def)
      end;
      writeln('    procedure set_'+member.name+'(const val: '+typestr+');');
      writeln('    function get_'+member.name+': '+typestr+';');
      last_name := member.name;
    end;
    writeln('    function discriminator: '+type_str(u.discriminator_type_def)+';');
  writeln('  end;');
  writeln;
  declare_marshaller(u);
end;

procedure TCodegen.code_declare_struct(str: IStructDef);
var
  members : StructMemberSeq;
  i : integer;
  contents : IInterfaceList;
begin
  members := nil;
  if FDB.repoid_defined((str as IContained).id) then exit;
  contents := (str as IContainer).contents();
  // internal declarations
  for i := 0 to Pred(contents.count) do
    impl_declare(contents[i] as IIRObject);
  members := str.members^;
  for i := 0 to Pred(Length(members)) do
    case members[i].type_def.def_kind of
      dk_Sequence : if not is_base_sequence(members[i].type_def as ISequenceDef) then
        code_declare_sequence(members[i].type_def as ISequenceDef);
      dk_Array : code_declare_array(members[i].type_def as IArrayDef);
    end;
  declare_marshaller(str);
end;

procedure TCodegen.code_declare_value(val: IValueDef);
var
  name, impl_name, int_name : string;
  mlist,list : IInterfaceList;
  i : integer;
  vmd: IValueMemberDef;
begin
  if FDB.repoid_defined((val as IContained).id) then exit;
  name := (val as IContained)._absolute_name;
  impl_name := 'T'+name;
  int_name := 'I'+name;
  writeln('  {** '+(val as IContained).id+' } ');
  writeln('  '+impl_name+' = class(TValueBase,'+int_name+',IValueBase)');
  mlist := (val as IContainer).contents(dk_ValueMember);
  writeln('  protected');
  for i := 0 to Pred(mlist.count) do
    begin
      vmd := mlist[i] as IValueMemberDef;
      writeln('    F'+name+'_'+vmd.name+' : '+type_str(vmd.get_type_def)+';');
    end;
  for i := 0 to Pred(mlist.count) do
    value_member(mlist[i] as IValueMemberDef);
  list := (val as IContainer).contents(dk_Operation);
  for i := 0 to Pred(list.count) do
    writeln('    '+operation_str(list[i] as IOperationDef)+' virtual; abstract;');
  writeln('    procedure _marshal_members(const enc: IEncoder); override;');
  writeln('    procedure _get_marshal_info(const str: TStrings;out chunked: Boolean); override;');
  writeln('    function  _demarshal_members(const dec: Idecoder): Boolean; override;');
  writeln('    function narrow_helper(const repoid: PChar): Pointer; override;');
  writeln('  public');
  writeln('    class function _downcast(const vb : IValueBase): '+int_name+';');
  writeln('  end;');
  writeln;
  // Init
  if Length(val.initializers) > 0 then
    begin
      writeln('  {** factory for valuetype '+(val as IContained).id+'} ');
      writeln('  '+impl_name+'_init = class(TInterfacedObject,'+int_name+'_init,IValueFactory)');
      writeln('  protected');
      for i := 0 to Pred(Length(val.initializers)) do
        writeln('    '+initializer_decl(val.initializers[i],val)+' virtual; abstract;');
      writeln('    function narrow_helper(const repoid: PChar): Pointer;');
      writeln('    function create_for_umarshal(): IValueBase; virtual; abstract;');
      writeln('  public');
      writeln('    class function _downcast(const vf : IValueFactory): '+int_name+'_init;');
      writeln('  end;');
      writeln;
    end;
  declare_marshaller(val);
end;

procedure TCodegen.code_impl_alias(cont: IAliasDef);
begin
  if FDB.repoid_defined((cont as IContained).id) then exit;
  alias_marshaller(cont);
end;

procedure TCodegen.code_impl_enum(enum: IEnumDef);
begin
  if FDB.repoid_defined((enum as IContained).id) then exit;
  enum_marshaller(enum as IContained);
end;

procedure TCodegen.final_impl_value(val: IValueDef);
var
  name,ostr : string;
  op: IOperationDef;
  list : IInterfaceList;
  i,ind : integer;
  vmd: IValueMemberDef;
begin
  if FDB.repoid_defined((val as IContained).id) then exit;
  name := (val as IContained)._absolute_name;
  writeln('//***********************************************************');
  writeln('// T'+name+'_impl');
  writeln('//***********************************************************');
  list := (val as IContainer).contents(dk_ValueMember);
  // constructor
  write('constructor T'+name+'_impl.Create(');
  for i := 0 to Pred(list.count) do
    begin
      if i > 0 then write('; ');
      vmd := list[i] as IValueMemberDef;
      write('_'+vmd.name+': '+type_str(vmd.get_type_def));
    end;
  writeln(');');
  writeln('begin');
  writeln('  // TODO');
  writeln('end;');
  writeln;
  // methods
  list := (val as IContainer).contents(dk_Operation);
  for i := 0 to Pred(list.count) do
    begin
      op := list[i] as IOperationDef;
      ostr := operation_str(op);
      ind := Pos(' ',ostr);
      writeln(Copy(ostr,1,ind)+'T'+name+'_impl.'+copy(ostr,ind+1,10000));
      writeln('begin');
      writeln('  // TODO');
      if op._result.kind <> tk_void then
        writeln('  result := '+def_value(op.get_result_def));
      writeln('end;');
      writeln;
    end;
  if Length(val.initializers) > 0 then
    begin
      writeln('//***********************************************************');
      writeln('// T'+name+'_factory');
      writeln('//***********************************************************');
      writeln('function T'+name+'_factory.create_for_umarshal(): IValueBase;');
      writeln('begin');
      write('  result := T'+name+'_impl.Create(');
      list := (val as IContainer).contents(dk_ValueMember);
      for i := 0 to Pred(list.count) do
        begin
          if i > 0 then write(', ');
          vmd := list[i] as IValueMemberDef;
          write(def_value(vmd.get_type_def));
        end;
      writeln(');');
      writeln('end;');
      writeln;
      for i := 0 to Pred(Length(val.initializers)) do
        begin
          ostr := initializer_decl(val.initializers[i],val);
          ind := Pos(' ',ostr);
          writeln(Copy(ostr,1,ind)+'T'+name+'_factory.'+copy(ostr,ind+1,10000));
          writeln('begin');
          writeln('  // TODO');
          writeln('  result := nil;');
          writeln('end;');
          writeln;
        end;
    end;
end;

procedure TCodegen.final_declare_value(val: IValueDef);
var
  name: string;
  list : IInterfaceList;
  i : integer;
  vmd: IValueMemberDef;
begin
  if FDB.repoid_defined((val as IContained).id) then exit;
  name := (val as IContained)._absolute_name;
  writeln('  T'+name+'_impl = class(T'+name+')');
  writeln('  protected');
  list := (val as IContainer).contents(dk_Operation);
  for i := 0 to Pred(list.count) do
    writeln('    '+operation_str(list[i] as IOperationDef)+' override;');
  writeln('  public');
  list := (val as IContainer).contents(dk_ValueMember);
  write('    constructor Create(');
  for i := 0 to Pred(list.count) do
    begin
      if i > 0 then write('; ');
      vmd := list[i] as IValueMemberDef;
      write('_'+vmd.name+': '+type_str(vmd.get_type_def));
    end;
  writeln(');');
  writeln('  end;');
  writeln;
  if Length(val.initializers) > 0 then
    begin
      writeln('  T'+name+'_factory = class(T'+name+'_init)');
      writeln('  protected');
      for i := 0 to Pred(Length(val.initializers)) do
        writeln('    '+initializer_decl(val.initializers[i],val)+' override;');
      writeln('    function create_for_umarshal(): IValueBase; override;');
      writeln('  end;');
      writeln;
    end;
end;

procedure TCodegen.code_impl_exception(e: IExceptionDef);
var
  members: StructMemberSeq;
  member : StructMember;
  i, mlen : integer;
  tstr,name,id: string;
begin
  members := nil;
  if FDB.repoid_defined(e.id) then exit;
  name := (e as IContained)._absolute_name;
  id := e.id;
  members := e.members^;
  mlen := Length(members);
  for i := 0 to Pred(mlen) do
    case members[i].type_def.def_kind of
      dk_Sequence : if not is_base_sequence(members[i].type_def as ISequenceDef) then
        seq_marshaller(members[i].type_def as ISequenceDef);
      dk_Array : array_marshaller(members[i].type_def as IArrayDef);
    end;
  except_marshaller(e);
  writeln('//***********************************************************');
  writeln('// T'+name);
  writeln('//***********************************************************');
  writeln('constructor T'+name+'.Create(ex: T'+name+');');
  writeln('begin');
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : tstr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : tstr := 'T'+name + '_'+member.name + '_array';
        else
          tstr := type_str(member.type_def)
      end;
      writeln('  F'+member.name+':= ex.'+member.name+';');
    end;
  writeln('  inherited Create();');
  writeln('end;');
  writeln;

  writeln('procedure T'+name+'.throw;');
  writeln('begin');
  writeln('  raise T'+name+'.Create(self);');
  writeln('end;');
  writeln;

  writeln('function T'+name+'.clone(): IORBException;');
  writeln('begin');
  writeln('  result := T'+name+'.Create(self);');
  writeln('end;');
  writeln;

  writeln('procedure T'+name+'.encode(const enc: IEncoder);');
  writeln('var');
  writeln('  int : I'+name+';');
  writeln('begin');
  writeln('  int := self;');
  writeln('  '+name+'_marshaller.marshal(enc,@int);');
  writeln('end;');
  writeln;

  write('constructor T'+name+'.Create(');
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : tstr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : continue;
        else
          tstr := type_str(member.type_def)
      end;
      write('_'+member.name+': '+ tstr);
      if i < mlen-1 then write(';');
    end;
  writeln(');');
  writeln('begin');
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      if member.type_def.def_kind <> dk_Array then
        writeln('  F'+member.name+' := _'+member.name+';');
    end;
  writeln('  inherited Create();');
  writeln('end;');
  writeln;
  writeln('function T'+name+'.repoid(): PChar;');
  writeln('begin');
  writeln('  result := '''+id+''';');
  writeln('end;');
  writeln;
  for i := 0 to Pred(mlen) do
    begin
      member := members[i];
      case member.type_def.def_kind of
        dk_Sequence : tstr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : tstr := 'T'+name + '_'+member.name + '_array';
        else
          tstr := type_str(member.type_def)
      end;
      writeln('procedure T'+name+'.'+member.name+'(const val: '+tstr+');');
      writeln('begin');
      writeln('  F'+member.name+' := val;');
      writeln('end;');
      writeln;
      writeln('function T'+name+'.'+member.name+': '+tstr+';');
      writeln('begin');
      writeln('  result := F'+member.name+';');
      writeln('end;');
      writeln;
    end;
end;

procedure TCodegen.code_impl_union(u: IUnionDef);
var
  i,ind : integer;
  members: UnionMemberSeq;
  member: UnionMember;
  typestr,name, val, last_name: string;
  discr: IIDLType;
  enum : IEnumDef;
  emembers : EnumMemberSeq;
begin
  members := nil;
  emembers := nil;
  if FDB.repoid_defined((u as IContained).id) then exit;
  name := (u as IContained)._absolute_name;
  members := u.members^;
  for i := 0 to Pred(Length(members)) do
    case members[i].type_def.def_kind of
      dk_Sequence : if not is_base_sequence(members[i].type_def as ISequenceDef) then
        seq_marshaller(members[i].type_def as ISequenceDef);
      dk_Array : array_marshaller(members[i].type_def as IArrayDef);
      dk_Struct : struct_marshaller(members[i].type_def as IContained);
    end;
  union_marshaller(u);
  writeln('//***********************************************************');
  writeln('// T'+name);
  writeln('//***********************************************************');
  discr := original_type(u.discriminator_type_def);
  if discr.def_kind = dk_Enum then
    begin
      enum := TEnumDef._narrow(discr as IORBObject);
      assert(enum <> nil);
      emembers := enum.members^;
    end
  else
    emembers := nil;
  last_name := '';
  for i := 0 to Pred(Length(members)) do
    begin
      member := members[i];
      if member.name = last_name then continue;
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : typestr := 'T'+name + '_'+member.name + '_array';
        else
          typestr := type_str(member.type_def)
      end;
      writeln('procedure T'+name+'.set_'+member.name+'(const val: '+typestr+');');
      writeln('begin');
      // default
      if member._label.get_type.kind = tk_octet then
        val := get_default_value_for_union(u.discriminator_type_def,members)
      else if discr.def_kind = dk_Enum then
        begin
          ind := member._label.get_variant;
          assert(ind < Length(emembers));
          val := emembers[ind];
        end
      else if discr._type.kind = tk_char then
        val := '''' + member._label.get_variant+ ''''
      else
        val := member._label.get_variant;
      writeln('  Fdiscriminator := '+val+';');
      writeln('  F'+member.name+' := val;');
      writeln('end;');
      writeln;
      writeln('function T'+name+'.get_'+member.name+': '+typestr+';');
      writeln('begin');
      writeln('  result := F'+member.name+';');
      writeln('end;');
      writeln;
      last_name := member.name;
    end;
  // discriminator
  writeln('function T'+name+'.discriminator: '+type_str(discr)+';');
  writeln('begin');
  writeln('  result := FDiscriminator;');
  writeln('end;');
  writeln;
end;

procedure TCodegen.code_impl_value(val: IValueDef);
var
  name, cname, tname, mname, fname : string;
  list: IInterfaceList;
  i : integer;
  vmd: IValueMemberDef;
  iseq: initializerseq;
begin
  iseq := nil;
  if FDB.repoid_defined((val as IContained).id) then exit;
  name := (val as IContained)._absolute_name;
  cname := 'T'+name;
  writeln('//***********************************************************');
  writeln('// '+cname);
  writeln('//***********************************************************');
  list := (val as IContainer).contents(dk_ValueMember);
  for i := 0 to Pred(list.count) do
    begin
      vmd := list[i] as IValueMemberDef;
      mname := vmd.name;
      tname := type_str(vmd.get_type_def);
      writeln('function '+cname+'.'+mname+': '+tname+';');
      writeln('begin');
      writeln('  result := F'+name+'_'+mname+';');
      writeln('end;');
      writeln;
      writeln('procedure '+cname+'.'+mname+'(val : '+tname+');');
      writeln('begin');
      writeln('  F'+name+'_'+mname+' := val;');
      writeln('end;');
      writeln;
    end;
  // narrow_helper
  writeln('function '+cname+'.narrow_helper(const repoid: PChar): Pointer;');
  writeln('begin');
  writeln('  result := nil;');
  writeln('  if strcomp(repoid,''IDL:'+name+':1.0'') = 0 then');
  writeln('    result := Pointer(Self as I'+name+');');
  writeln('end;');
  writeln;
  // _downcast
  writeln('class function '+cname+'._downcast(const vb : IValueBase): I'+name+';');
  writeln('var');
  writeln('  ptr: pointer;');
  writeln('begin');
  writeln('  result := nil;');
  writeln('  if vb <> nil then');
  writeln('    begin');
  writeln('      ptr := vb.narrow_helper(''IDL:'+name+':1.0'');');
  writeln('      if ptr <> nil then');
  writeln('        result := I'+name+'(ptr);');
  writeln('    end;');
  writeln('end;');
  writeln;
  // _get_marshal_info
  writeln('procedure '+cname+'._get_marshal_info(const str: TStrings; out chunked: Boolean);');
  writeln('begin');
  writeln('  str.Add(''IDL:'+name+':1.0'');');
  writeln('  chunked := false;');
  writeln('end;');
  writeln;
  // _marshal_members
  writeln('procedure '+cname+'._marshal_members(const enc: IEncoder);');
  writeln('begin');
  for i := 0 to Pred(list.count) do
    begin
      vmd := list[i] as IValueMemberDef;
      mname := vmd.name;
      writeln('  '+marshaller_for(vmd.get_type_def)+'.marshal(enc,@F'+name+'_'+mname+');');
    end;
  writeln('end;');
  writeln;
  // _demarshal_members
  writeln('function  '+cname+'._demarshal_members(const dec: IDecoder): Boolean;');
  writeln('begin');
  writeln('  result := false;');
  for i := 0 to Pred(list.count) do
    begin
      vmd := list[i] as IValueMemberDef;
      mname := vmd.name;
      writeln('  if not '+marshaller_for(vmd.get_type_def)+'.demarshal(dec,@F'+name+'_'+mname+') then exit;');
    end;
  writeln('  result := true;');
  writeln('end;');
  writeln;
  // init
  iseq := val.initializers;
  if Length(iseq) > 0 then
    begin
      fname := cname+'_init';
      writeln('//***********************************************************');
      writeln('// '+fname);
      writeln('//***********************************************************');
      // _downcast
      writeln('class function '+fname+'._downcast(const vf : IValueFactory): I'+name+'_init;');
      writeln('var');
      writeln('  ptr: pointer;');
      writeln('begin');
      writeln('  result := nil;');
      writeln('  if vf <> nil then');
      writeln('    begin');
      writeln('      ptr := vf.narrow_helper(''IDL:'+name+':1.0'');');
      writeln('      if ptr <> nil then result := I'+name+'_init(ptr);');
      writeln('    end;');
      writeln('end;');
      writeln;
      // narrow_helper
      writeln('function '+fname+'.narrow_helper(const repoid: PChar): Pointer;');
      writeln('begin');
      writeln('  result := nil;');
      writeln('  if strcomp(repoid,''IDL:'+name+':1.0'')= 0 then');
      writeln('    result := Pointer( self as I'+name+'_init);');
      writeln('end;');
      writeln;
    end;
  // Marshaller
  mname := cname+'_marshaller';
  writeln('//***********************************************************');
  writeln('// '+mname);
  writeln('//***********************************************************');
  // _create
  writeln('procedure '+mname+'._create(var addr: Pointer);');
  writeln('begin');
  writeln('  I'+name+'(addr^) := I'+name+'('+cname+'.Create);');
  writeln('end;');
  writeln;
  // _assign
  writeln('procedure '+mname+'._assign(dst, src: Pointer);');
  writeln('begin');
  writeln('  I'+name+'(dst^) := I'+name+'(src^);');
  writeln('end;');
  writeln;
  // _free
  writeln('procedure '+mname+'._free(var addr: Pointer);');
  writeln('begin');
  writeln('  I'+name+'(addr^) := nil;');
  writeln('end;');
  writeln;
  //  demarshal
  writeln('function '+mname+'.demarshal(dec: IDecoder;');
  writeln('  const addr: Pointer): Boolean;');
  writeln('var');
  writeln('  vb : IValueBase;');
  writeln('begin');
  writeln('  result := TValueBase._demarshal(dec,vb,''IDL:'+name+':1.0'');');
  writeln('  if not result then exit;');
  writeln('  IValueBase(addr^) := '+cname+'._downcast(vb);');
  writeln('  result := Pointer(addr^) <> nil;');
  writeln('end;');
  writeln;
  //  marshal
  writeln('procedure '+mname+'.marshal(enc: IEncoder; const addr: Pointer);');
  writeln('begin');
  writeln('  TValueBase._marshal(enc,IValueBase(addr^));');
  writeln('end;');
  writeln;
  if FParams.IndexOf('-any')>=0 then
    begin
      //  typecode
      writeln('function '+mname+'.typecode: ITypeCode;');
      writeln('begin');
      writeln('  result := _tc_'+name+'.typecode;');
      writeln('end;');
      writeln;
    end;
end;

procedure TCodegen.code_impl_module(m: IModuleDef);
var
  i : integer;
  contents : IInterfaceList;
begin
  //if FDB.repoid_defined(m.id) then exit;
  contents := (m as IContainer).contents();
  for i := 0 to Pred(contents.count) do
    impl_code(contents[i] as IContained);
end;

procedure TCodegen.code_impl_interface(int: IInterfaceDef);
var
  id_str, name, cname, iname, biname: string;
  bid_str, stub_name, skel_name: string;
  base, methods, bmethods, attributes: IInterfaceList;
  baseint: IInterfaceDef;
  i, j: integer;
  contents: IInterfaceList;

  procedure impl_stub_methods(int: IInterfaceDef; except_first: boolean);
  var
    i : integer;
    list: IInterfaceList;
    n: integer;
  begin
    list := int.get_base_interfaces;
    if except_first then n := 1 else n := 0;
    for i := n to Pred(list.count) do
      impl_stub_methods(list[i] as IInterfaceDef, except_first and (i = n));
    list := (int as IContainer).contents(dk_Operation);
    for i := 0 to Pred(list.count) do
      stub_method(name,list[i] as IOperationDef);
    list := (int as IContainer).contents(dk_Attribute);
    for i := 0 to Pred(list.count) do
      stub_attribute(name,list[i] as IAttributeDef);
  end;

begin
  if FDB.repoid_defined(int.id) then exit;
  contents := (int as IContainer).contents();
  // internal declarations
  for i := 0 to Pred(contents.count) do
    impl_code(contents[i] as IContained);
  name := IContained(int)._absolute_name;
  cname := 'T' + name;
  stub_name := cname + '_stub';
  skel_name := cname + '_skel';
  iname := 'I' + name;
  writeln('//***********************************************************');
  writeln('// '+cname);
  writeln('//***********************************************************');
  methods := (int as IContainer).contents(dk_Operation);
  attributes := (int as IContainer).contents(dk_Attribute);
  base := int.get_base_interfaces;
  // narrow_helper
  writeln('function '+cname+'.narrow_helper(const str: string): Pointer;');
  writeln('begin');
  id_str := ''''+int.id+'''';
  if base.count = 0 then
    writeln('  result := nil;');
  writeln('  if str = '+id_str+' then');
  writeln('    begin');
  writeln('      result := Pointer(self as '+iname+');');
  writeln('      exit;');
  writeln('    end;');
  for i := 1 to Pred(base.count) do
    begin
      baseint := base[i] as IInterfaceDef;
      bid_str := ''''+baseint.id+'''';
      iname := 'I' + baseint._absolute_name;
      writeln('  if str = '+bid_str+' then');
      writeln('    begin');
      writeln('      result := Pointer(self as '+iname+');');
      writeln('      exit;');
      writeln('    end;');
    end;
  if base.count > 0 then
    writeln('  result := inherited narrow_helper(str);');
  writeln('end;');
  writeln;
  // _narrow
  iname := 'I' + name;
  writeln('class function '+cname+'._narrow(const obj : IORBObject): '+iname+';');
  writeln('var');
  writeln('  p: Pointer;');
  writeln('  stub: '+stub_name+';');
  writeln('begin');
  writeln('  result := nil;');
  writeln('  if obj = nil then exit;');
  writeln('  p := obj.narrow_helper('+id_str+');');
  writeln('  if p <> nil then');
  writeln('    result := '+iname+'(p)');
  writeln('  else');
  writeln('    begin');
  writeln('      if narrow_helper2(obj) or obj._is_a_remote('+id_str+') then');
  writeln('        begin');
  writeln('          stub := '+stub_name+'.Create();');
  writeln('          stub.assign(obj);');
  writeln('          result := stub;');
  writeln('        end;');
  writeln('    end;');
  writeln('end;');
  writeln;
  // narrow_helper2
  writeln('class function '+cname+'.narrow_helper2(const obj: IORBObject): Boolean;');
  writeln('begin');
  writeln('  result := obj.repoid = '+id_str);
  writeln('end;');
  writeln;

  // stub methods
  writeln('//***********************************************************');
  writeln('// '+stub_name);
  writeln('//***********************************************************');
  impl_stub_methods(int, false);
  if FParams.IndexOf('-poa') < 0 then
    begin
      // BOA skel
      writeln('//***********************************************************');
      writeln('// '+skel_name);
      writeln('//***********************************************************');
      writeln('constructor '+skel_name+'.Create;');
      writeln('var');
      writeln('  impl: IImplementationDef;');
      writeln('begin');
      writeln('  inherited Create();');
      writeln('  impl := find_impl('+id_str+','''+name+''');');
      writeln('  create_ref('''',nil,impl,'+id_str+');');
      for j := 1 to Pred(base.count) do
        begin
          baseint := base[j] as IInterfaceDef;
          writeln('  impl := find_impl('''+baseint.id+''','''+baseint._absolute_name+''');');
          writeln('  create_ref('''',nil,impl,'''+baseint.id+''');');
        end;
      writeln('  register_dispatcher(IStaticInterfaceDispatcher(self));');
      writeln('end;');
      writeln;
      method_dispatch(skel_name,int,true);
      writeln('function '+skel_name+'.narrow_helper(const str: string): Pointer;');
      writeln('begin');
      if base.count = 0 then
        writeln('  result := nil;');
      writeln('  if str = '+id_str+' then');
      writeln('    begin');
      writeln('      result := Pointer(self as I'+name+');');
      writeln('      exit;');
      writeln('    end;');
      for i := 1 to Pred(base.count) do
        begin
          baseint := base[i] as IInterfaceDef;
          bid_str := ''''+baseint.id+'''';
          iname := 'I' + baseint._absolute_name;
          writeln('  if str = '+bid_str+' then');
          writeln('    begin');
          writeln('      result := Pointer(self as '+iname+');');
          writeln('      exit;');
          writeln('    end;');
        end;
      if base.count > 0 then
        begin
          writeln('  result := inherited narrow_helper(str);');
        end;
      writeln('end;');
      writeln;
    end
  else
    begin
      // POA stub
      writeln('//***********************************************************');
      writeln('// '+cname+'_stub_clp');
      writeln('//***********************************************************');
      // narrow_helper
      writeln('function '+cname+'_stub_clp.narrow_helper(const str: string): Pointer;');
      writeln('begin');
      writeln('  result := nil;');
      writeln('  if str = '+id_str+' then');
      writeln('    begin');
      writeln('      result := Pointer(self as '+iname+');');
      writeln('      exit;');
      writeln('    end;');
      writeln('end;');
      writeln;
      for i := 0 to Pred(methods.count) do
        stub_clp_method(name,name,methods[i] as IOperationDef);
      for j := 1 to Pred(base.count) do
        begin
          bmethods := (base[j] as IContainer).contents(dk_Operation);
          for i := 0 to Pred(bmethods.count) do
            stub_clp_method(name,(base[j] as IInterfaceDef)._absolute_name,bmethods[i] as IOperationDef);
        end;
      for i := 0 to Pred(attributes.Count) do
        stub_clp_attribute(name,attributes[i] as IAttributeDef);
      // POA servant
      writeln('//***********************************************************');
      writeln('// '+cname+'_serv');
      writeln('//***********************************************************');
      if FParams.IndexOf('-notabs') >= 0 then begin
        // Create
        writeln('constructor '+cname+'_serv.Create(const AImpl: '+iname+');');
        writeln('begin');
        if base.count = 0 then
          writeln('  inherited Create();')
        else
          writeln('  inherited Create(AImpl);');
        writeln('  FImplementation := AImpl;');
        writeln('end;');
        writeln;
        // methods
        for i := 0 to Pred(methods.count) do
          serv_method(name,methods[i] as IOperationDef);
        //get and set attributes methods
        for i := 0 to Pred(attributes.Count) do
          serv_attribute(name, attributes[i] as IAttributeDef);
      end;
      // _this
      writeln('function '+cname+'_serv._this(): '+iname+';');
      writeln('var');
      writeln('  obj: IORBObject;');
      writeln('begin');
      writeln('  obj := this();');
      writeln('  result := '+cname+'._narrow(obj);');
      writeln('end;');
      writeln;
      // _is_a
      writeln('function '+cname+'_serv._is_a(const repoid: string): Boolean;');
      writeln('begin');
      writeln('  result := repoid = '+id_str+';');
      writeln('end;');
      writeln;
      // _primary_interface
      writeln('function '+cname+'_serv._primary_interface(const objid: ObjectID; const poa: IPOA): string;');
      writeln('begin');
      writeln('  result := '+id_str+';');
      writeln('end;');
      writeln;
      // _make_stub
      writeln('function '+cname+'_serv._make_stub(const poa: IPOA; const obj : IORBObject): IORBObject;');
      writeln('begin');
      writeln('  result := '+cname+'_stub_clp.Create(poa,obj);');
      writeln('end;');
      writeln;
      writeln('procedure '+cname+'_serv.invoke(const serv: IStaticServerRequest);');
      writeln('begin');
      writeln('  if not _dispatch(serv) then');
      writeln('    begin');
      writeln('      serv.set_exception(SystemException.Create_type(st_BAD_OPERATION,0,COMPLETED_NO) as IORBException);');
      writeln('      serv.write_results();');
      writeln('    end;');
      writeln('end;');
      writeln;
      // narrow_helper
      writeln('function '+cname+'_serv.narrow_helper(const str: string): Pointer;');
      writeln('begin');
      if base.count = 0 then
        writeln('  result := nil;');
      writeln('  if str = '+id_str+' then');
      writeln('    begin');
      writeln('      result := Pointer('+iname+'(self));');
      writeln('      exit;');
      writeln('    end;');
      base := int.get_base_interfaces;
      for i := 1 to Pred(base.count) do
        begin
          baseint := base[i] as IInterfaceDef;
          bid_str := ''''+baseint.id+'''';
          biname := 'I' + baseint._absolute_name;
          writeln('  if str = '+bid_str+' then');
          writeln('    begin');
          writeln('      result := Pointer(self as '+biname+');');
          writeln('      exit;');
          writeln('    end;');
        end;
      if base.count > 0 then
        begin
          writeln('  result := inherited narrow_helper(str);');
        end;
      writeln('end;');
      writeln;
      // _narrow
      writeln('class function '+cname+'_serv._narrow(const srv : IServant): '+iname+';');
      writeln('var');
      writeln('  p: Pointer;');
      writeln('begin');
      writeln('  result := nil;');
      writeln('  if srv = nil then exit;');
      writeln('  p := srv.narrow_helper('+id_str+');');
      writeln('  if p <> nil then');
      writeln('    result := '+iname+'(p)');
      writeln('end;');
      writeln;
      method_dispatch(cname+'_serv',int,false);
    end;
  //----------- marshaller ---------------
  int_marshaller(int);
end;


procedure TCodegen.code_impl_struct(str: IStructDef);
var
  i : integer;
  members : StructMemberSeq;
  contents : IInterfaceList;
begin
  members := nil;
  if FDB.repoid_defined((str as IContained).id) then exit;
  contents := (str as IContainer).contents();
  // internal declarations
  for i := 0 to Pred(contents.count) do
    impl_code(contents[i] as IContained);
  members := str.members^;
  for i := 0 to Pred(Length(members)) do
    case members[i].type_def.def_kind of
      dk_Sequence : if not is_base_sequence(members[i].type_def as ISequenceDef) then
        seq_marshaller(members[i].type_def as ISequenceDef);
      dk_Array : array_marshaller(members[i].type_def as IArrayDef);
    end;
  struct_marshaller(str as IContained)
end;

procedure TCodegen.int_declare_sequence(seq: ISequenceDef);
var
  tname : string;
  bound : longint;
begin
  //if FDB.repoid_defined(FDB.pseudo_repoid(seq)) then  exit;
  tname := type_str(seq.element_type_def);
  bound := seq.bound;
  if bound = 0 then
    writeln('array of '+ tname+';')
  else
    writeln('array[0..'+inttostr(bound-1)+'] of '+ tname+';')
end;

procedure TCodegen.int_declare_array(arr: IArrayDef);
begin
  if FDB.repoid_defined(FDB.pseudo_repoid(arr)) then  exit;
  writeln(array_str(arr));
end;

procedure TCodegen.int_declare_alias(cont: IAliasDef);
var
  c : IContained;
  t : IIDLType;
begin
  if FDB.repoid_defined((cont as IContained).id) then exit;
  c := cont as IContained;
  write('  T'+ c._absolute_name +' = ');
  t :=  cont.original_type_def;
  case t.def_kind of
    dk_Sequence : int_declare_sequence(t as ISequenceDef);
    dk_Enum : writeln('T' + (t as IContained)._absolute_name+';');
    dk_Primitive : writeln(type_str(t)+';');
    dk_Struct: writeln(type_str(t)+';');
    dk_Array : writeln(array_str(t as IArrayDef));
    dk_Alias : writeln('T'+(t as IContained)._absolute_name+';');
    dk_Interface : begin
        if FDB.repoid_defined((t as IContained).id) then
          writeln(ChangeFileExt(FDB.repoid_file_name((t as IContained).id),'_int')+'.I'+(t as IContained).name)
        else
          writeln('I'+(t as IContained)._absolute_name+';');
      end;
    else
      assert(false,'not implemented');
  end;
  writeln;
end;

procedure TCodegen.int_declare_enum(enum: IEnumDef);
var
  i : integer;
  name : string;
begin
  if FDB.repoid_defined((enum as IContained).id) then exit;
  name := (enum as IContained)._absolute_name;
  if not declare(name) then exit;
  writeln('  T'+name+' = (');
  for i := 0 to Pred(length(enum.members^)) do
    if i = Pred(length(enum.members^)) then
      writeln('    '+enum.members^[i])
    else
      writeln('    '+enum.members^[i]+',');
  writeln('  );');
  writeln;
end;

procedure TCodegen.int_declare_module(m: IModuleDef);
var
  i : integer;
  contents : IInterfaceList;
begin
  if not FDB.repoid_defined(m.id) then begin
    writeln('// module '+m.name+' types');
    writeln;
  end;
  contents := (m as IContainer).contents();
  for i := 0 to Pred(contents.count) do
    int_declare(contents[i] as IContained);
end;

procedure TCodegen.int_declare_internal_type(cont_name,name: string; t: IIDLType);
var
  tmp : string;
begin
  case t.def_kind of
    dk_Union : int_declare_union(t as IUnionDef);
    dk_Enum : int_declare_enum(t as IEnumDef);
    dk_Struct: int_declare_struct(t as IStructDef);
    dk_Sequence: begin
        tmp := cont_name+'_'+name+'_seq';
        if not declare(tmp) then exit;
        write('  T'+tmp+' = ');
        int_declare_sequence(t as ISequenceDef);
        writeln;
      end;
    dk_Array: begin
        tmp := cont_name+'_'+name+'_array';
        if not declare(tmp) then exit;
        write('  T'+tmp+' = ');
        int_declare_array(t as IArrayDef);
        writeln;
      end;
  end;
end;

procedure TCodegen.int_declare_union(u: IUnionDef);
var
  members : UnionMemberSeq;
  member: UnionMember;
  name,typestr,last_name : string;
  i : integer;
{$IFDEF LINUX}
  newGUID: TGUID;
{$ENDIF}

  procedure internal_types;
  var
    i : integer;
    name: string;
  begin
    name := (u as IContained)._absolute_name;
    members := u.members^;
    for i := 0 to Pred(Length(members)) do
      int_declare_internal_type(name,members[i].name,members[i].type_def);
  end;

begin
  if FDB.repoid_defined((u as IContained).id) then exit;
  name := (u as IContained)._absolute_name;
  if not declare(name) then exit;
  internal_types;
  members := u.members^;
  writeln('  I'+name+' = interface');
{$IFDEF MSWINDOWS}
  writeln('  ['''+CreateClassID()+''']');
{$ENDIF}
{$IFDEF LINUX}
  CreateGUID(newGUID);
  writeln('  ['''+GUIDToString(newGUID)+''']');
{$ENDIF}
  for i := 0 to Pred(Length(members)) do
    begin
      member := members[i];
      if member.name = last_name then continue;
      case member.type_def.def_kind of
        dk_Sequence : typestr := 'T'+name + '_'+member.name + '_seq';
        dk_Array : typestr := 'T'+name + '_'+member.name + '_array';
        else
          typestr := type_str(member.type_def)
      end;
      writeln('    procedure set_'+member.name+'(const val: '+typestr+');');
      writeln('    function get_'+member.name+': '+typestr+';');
      last_name := member.name;
    end;
    writeln('    function discriminator: '+type_str(u.discriminator_type_def)+';');
  writeln('  end;');
  writeln;
end;

procedure TCodegen.int_declare_interface(int: IInterfaceDef);
var
  iname,aname : string;
  base_int,methods,attributes : IInterfaceList;
  i : integer;
  contents : IInterfaceList;
  attr : IAttributeDef;
{$IFDEF LINUX}
  newGUID: TGUID;
{$ENDIF}
begin
  if FDB.repoid_defined(int.id) then exit;
  if not declare(IContained(int)._absolute_name) then exit;
  contents := (int as IContainer).contents();
  // internal declarations
  for i := 0 to Pred(contents.count) do
    int_declare(contents[i] as IContained);
  iname := 'I'+IContained(int)._absolute_name;
  write('  '+iname+' = interface');
  base_int := int.get_base_interfaces;
  if base_int.Count > 0 then
    writeln('(I'+IContained(base_int[0])._absolute_name+')')
  else
    writeln;
{$IFDEF MSWINDOWS}
  writeln('  ['''+CreateClassID()+''']');
{$ENDIF}
{$IFDEF LINUX}
  CreateGUID(newGUID);
  writeln('  ['''+GUIDToString(newGUID)+''']');
{$ENDIF}
  methods := (int as IContainer).contents(dk_Operation);
  for i := 0 to Pred(methods.count) do
    writeln('    '+operation_str(methods[i] as IOperationDef));
  attributes := (int as IContainer).contents(dk_Attribute);
  for i := 0 to Pred(attributes.count) do
    begin
      attr:= attributes[i] as IAttributeDef;
      if attr.get_mode = ATTR_NORMAL then
        writeln('    '+attribute_str(attr,true));
      writeln('    '+attribute_str(attr,false));
    end;
  for i := 0 to Pred(attributes.count) do
    begin
      attr:= attributes[i] as IAttributeDef;
      aname := attr.name;
      write('    property '+aname+': '+ type_str(attr.get_type_def)+' read _get_'+aname);
      if attr.get_mode = ATTR_NORMAL then
        writeln(' write _set_'+aname+';')
      else
        writeln(';');
    end;
  writeln('  end;');
  writeln;
end;

procedure TCodegen.int_declare_exception(e : IExceptionDef);
var
  ename: string;
  members: StructMemberSeq;
  i : integer;
{$IFDEF LINUX}
  newGUID: TGUID;
{$ENDIF}
begin
  members := nil;
  if FDB.repoid_defined(e.id) then exit;
  ename := safe_name((e as IContained)._absolute_name);
  if not declare(ename) then exit;
  members := e.members^;
  // declare internally defined types
  for i := 0 to Pred(Length(members)) do
    int_declare_internal_type(ename,members[i].name,members[i].type_def);
  writeln('  I'+ename+' = interface(IUserException)');
{$IFDEF MSWINDOWS}
  writeln('  ['''+CreateClassID()+''']');
{$ENDIF}
{$IFDEF LINUX}
  CreateGUID(newGUID);
  writeln('  ['''+GUIDToString(newGUID)+''']');
{$ENDIF}
  ex_members_decl(ename,members);
  writeln('  end;');
  writeln
end;

procedure TCodegen.int_declare_struct(str: IStructDef);
var
  sname,tstr: string;
  i : integer;
  members : StructMemberSeq;
begin
  members := nil;
  if FDB.repoid_defined((str as IContained).id) then exit;
  sname := safe_name((str as IContained)._absolute_name);
  if not declare(sname) then exit;
  members := str.members^;
  // internal types
  for i := 0 to Pred(Length(members)) do
    int_declare_internal_type(sname,members[i].name,members[i].type_def);
  writeln('  T'+sname+' = record');
  for i := 0 to Pred(Length(members)) do
    begin
      case members[i].type_def.def_kind of
        dk_Sequence : tstr := 'T'+sname+'_'+members[i].name+'_seq';
        dk_Array : tstr := 'T'+sname+'_'+members[i].name+'_array';
        else
          tstr := type_str(members[i].type_def);
      end;
      writeln('    '+safe_name(members[i].name)+' : '+tstr+';');
    end;
  writeln('  end;');
  writeln;
end;

procedure TCodegen.int_declare_value(val: IValueDef);
var
  vname : string;
  list : IInterfaceList;
  i : integer;
{$IFDEF LINUX}
  newGUID: TGUID;
{$ENDIF}
begin
  if FDB.repoid_defined((val as IContained).id) then exit;
  list := (val as IContainer).contents();
  vname := 'I'+(val as IContained)._absolute_name;
  if not declare(vname) then exit;
  // internal declarations
  for i := 0 to Pred(list.count) do
    int_declare(list[i] as IContained);
  writeln('  '+vname+' = interface(IValueBase)');
{$IFDEF MSWINDOWS}
  writeln('  ['''+CreateClassID()+''']');
{$ENDIF}
{$IFDEF LINUX}
  CreateGUID(newGUID);
  writeln('  ['''+GUIDToString(newGUID)+''']');
{$ENDIF}
  list := (val as IContainer).contents(dk_Operation);
  for i := 0 to Pred(list.count) do
    writeln('    '+operation_str(list[i] as IOperationDef));
  list := (val as IContainer).contents(dk_ValueMember);
  for i := 0 to Pred(list.count) do
    value_member(list[i] as IValueMemberDef);
  writeln('  end;');
  writeln;
  // Init interface
  if Length(val.initializers) > 0 then
    begin
      writeln('  '+vname+'_init = interface(IValueFactory)');
{$IFDEF MSWINDOWS}
      writeln('  ['''+CreateClassID()+''']');
{$ENDIF}
{$IFDEF LINUX}
      CreateGUID(newGUID);
      writeln('  ['''+GUIDToString(newGUID)+''']');
{$ENDIF}
      for i := 0 to Pred(Length(val.initializers)) do
        writeln('    '+initializer_decl(val.initializers[i],val));
      writeln('  end;');
      writeln;
    end;
end;

procedure TCodegen.final_code(cont: IContained);
var
  i : integer;
  contents : IInterfaceList;
begin
  case IIRObject(cont).def_kind of
    dk_Interface : final_impl_interface(IInterfaceDef(cont));
    dk_Value : final_impl_value(cont as IValueDef);
    dk_Module : begin
        contents := (cont as IContainer).contents;
        for i := 0 to Pred(contents.count) do
          final_code(contents[i] as IContained);
      end;
  end;
end;

procedure TCodegen.final_declare(cont: IContained);
var
  i : integer;
  contents: IInterfaceList;
begin
  case IIRObject(cont).def_kind of
    dk_Interface : final_declare_interface(IInterfaceDef(cont));
    dk_Value : final_declare_value(cont as IValueDef);
    dk_Module : begin
        contents := (cont as IContainer).contents;
        for i := 0 to Pred(contents.count) do
          final_declare(contents[i] as IContained);
      end;
  end;
end;

procedure TCodegen.final_declare_interface(int: IInterfaceDef);
var
  name : string;
begin
  if FDB.repoid_defined(int.id) then exit;
  name := IContained(int)._absolute_name;
  if FParams.IndexOf('-poa') < 0 then
    writeln('  T'+name+'_impl = class(T'+name+'_skel)')
  else
    writeln('  T'+name+'_impl = class(T'+name+'_serv)');
  declare_all_methods(int,' override;',false);
  writeln('  end;');
  writeln;
end;

procedure TCodegen.final_impl_interface(int: IInterfaceDef);
var
  name : string;
  base,methods,attributes : IInterfaceList;


  procedure final_method(def: IOperationDef);
  var
    mstr: string;
    ind: integer;
    res: IIDLType;
    i : integer;
  begin
    mstr := operation_str(def);
    ind := Pos(' ',mstr);
    writeln(copy(mstr,1,ind)+'T'+name+'_impl.'+copy(mstr,ind+1,10000));
    writeln('begin');
    writeln('  // TODO');
    res := def.get_result_def;
    while res.def_kind = dk_alias do
      res := (res as IAliasDef).original_type_def;
    if not (res._type.kind in [tk_void,tk_struct,tk_array]) then
      writeln('  result := '+def_value(def.get_result_def)+';');
    for i := 0 to Pred(Length(def.exceptions^)) do
      begin
        writeln('  dorb_throw(T'+def.exceptions^[i]._absolute_name+'.Create() as IORBException);');
      end;
    writeln('end;');
    writeln;
  end;

  procedure final_attribute(attr: IAttributeDef);
  var
    tstr: string;
  begin
    tstr := type_str(attr.get_type_def);
    if attr.get_mode = ATTR_NORMAL then
      begin
        writeln('procedure T'+name+'_impl._set_'+attr.name+'(const val: '+tstr+');');
        writeln('begin');
        writeln('  // TODO');
        writeln('end;');
        writeln;
      end;
    writeln('function T'+name+'_impl._get_'+attr.name+': '+tstr+';');
    writeln('begin');
    writeln('  // TODO');
    writeln('  result := '+def_value(attr.get_type_def)+';');
    writeln('end;');
    writeln;
  end;

  procedure final_methods(int: IInterfaceDef);
  var
    i : integer;
  begin
    base := int.get_base_interfaces;
    for i := 0 to Pred(base.count) do
      final_methods(base[i] as IInterfaceDef);
    methods := (int as IContainer).contents(dk_Operation);
    attributes := (int as IContainer).contents(dk_Attribute);
    for i := 0 to Pred(methods.Count) do
      final_method(methods[i] as IOperationDef);
    for i := 0 to Pred(attributes.Count) do
      final_attribute(attributes[i] as IAttributeDef);
  end;

begin
  if FDB.repoid_defined(int.id) then exit;
  name := IContained(int)._absolute_name;
  writeln('//***********************************************************');
  writeln('// '+name+'_impl');
  writeln('//***********************************************************');
  final_methods(int);
end;

procedure TCodegen.collect_constant(con: IConstantDef);
begin
  FConstants.add(con);
end;

procedure TCodegen.declare_constant(con: IConstantDef);
type
  ds = record
    case Boolean of
      true: (d: double );
      false: (arr: array[0..7] of byte);
  end;
  fs = record
    case Boolean of
      true: (f: float );
      false: (arr: array[0..3] of byte);
  end;
var
  name : string;
  val: Variant;
  ch: string;
begin
  if FDB.repoid_defined(con.id) then Exit;
  name := (con as IContained)._absolute_name;
  val := con.value.get_variant;
  case con.type_def._type.kind of
    tk_string: writeln('  '+name+' = '''+val+''';');
    tk_char: begin
        ch := val;
        if ord(ch[1]) >= 33 then
          writeln('  '+name+' = '''+ch+''';')
        else
          writeln('  '+name+' = #'+IntToStr(ord(char(ch[1])))+';')
      end;
    tk_short,tk_long,tk_ushort,tk_ulong: writeln('  '+name+' = '+IntToStr(val)+';');
    tk_float : writeln('  '+name+' = '+FloatToStrf(val,ffGeneral,7,0)+';');
    tk_double : writeln('  '+name+' = '+FloatToStr(val)+';');
    tk_boolean: if val = TRUE then
      writeln('  '+name+' = TRUE;')
    else
      writeln('  '+name+' = FALSE;');
    else
      assert(false,'not_implemented');
  end;
end;

function TCodegen.get_default_value_for_union(def: IIDLType;
  members: UnionMemberSeq): Variant;
var
  tc: ITypeCode;
  discr : IIDLType;
  enum: IEnumDef;
  i : integer;
  emembers: EnumMemberSeq;

  function in_use(val: Variant): Boolean;
  var
    j : integer;
  begin
    result := false;
    for j := 0 to Pred(Length(members)) do
      if (members[j]._label.get_type.kind <> tk_octet) and (members[j]._label.get_variant = val) then
        begin
          result := true;
          exit;
        end;
  end;
begin
  emembers := nil;
  tc := def._type;
  case tc.unalias.kind of
    tk_enum: begin
        discr := original_type(def);
        enum := TEnumDef._narrow(discr as IORBObject);
        assert(enum <> nil);
        emembers := enum.members^;
        for i := 0 to Pred(Length(emembers)) do
          if not in_use(i) then
            begin
              result := emembers[i];
              exit;
            end;
      end;
    tk_short,tk_long,tk_ushort,tk_ulong: begin
        for i := 0 to High(short) do
          if not in_use(i) then
            begin
              result := i;
              exit;
            end;
      end;
    tk_char : begin
        for i := 0 to High(Byte) do
          if not in_use(char(i)) then
            begin
              result := 'Char('+inttostr(i)+')';
              exit;
            end;
      end;
    else
      assert(false,'not implemented');
  end;
  raise Exception.Create('no values for default case');
end;

procedure TCodegen.union_dependencies(u: IUnionDef);
var
  i : integer;
  members : UnionMemberSeq;
begin
  members := u.members^;
  for i := 0 to Pred(Length(members)) do
    case members[i].type_def.def_kind of
      dk_Interface: add_forwards(members[i].type_def as IContained);
      dk_Struct,dk_Union,dk_Sequence,dk_Array :
        cont_dependencies(members[i].type_def);
    end;
  declare((u as IContained)._absolute_name);
end;

procedure TCodegen.struct_dependencies(val: IStructDef);
var
  i : integer;
  members : StructMemberSeq;
begin
  members := val.members^;
  for i := 0 to Pred(Length(members)) do
    case members[i].type_def.def_kind of
      dk_Interface: add_forwards(members[i].type_def as IContained);
      dk_Struct,dk_Union,dk_Sequence,dk_Array :
        cont_dependencies(members[i].type_def);
    end;
  declare((val as IContained)._absolute_name);
end;

procedure TCodegen.array_dependencies(a: IArrayDef);
begin
  with a.element_type_def do
    case def_kind of
      dk_Interface : add_forwards(a.element_type_def as IContained);
      dk_Struct : add_forwards(a.element_type_def as IContained);
    end;
end;

procedure TCodegen.seq_dependencies(s: ISequenceDef);
begin
  with s.element_type_def do
    case def_kind of
      dk_Interface : add_forwards(s.element_type_def as IContained);
      dk_Struct : add_forwards(s.element_type_def as IContained);
    end;
end;

procedure TCodegen.declare_forwards;
var
  i : integer;
begin
  for i := 0 to Pred(FForwards.count) do
    with IContained(FForwards[i]) do
      case def_kind of
        //dk_Interface,dk_Union : writeln('  I'+_absolute_name+ ' = interface;'+ char(CR));
        dk_Interface,dk_Union : begin writeln('  I'+_absolute_name+ ' = interface;'); writeln; end;
        //dk_Struct : writeln('  T'+_absolute_name+ ' = record;'+ char(CR));
        dk_Struct : begin writeln('  T'+_absolute_name+ ' = record;'); writeln; end;
      end;
end;

procedure TCodegen.attr_dependencies(attr: IAttributeDef);
var
  cont : IContained;
begin
  if attr._type.kind in [tk_objref,tk_struct,tk_union] then
    begin
      cont := TContained._narrow(attr.get_type_def as IORBObject);
      if (cont <> nil) and (FDeclared.indexOf(cont._absolute_name) < 0) then
        add_forwards(cont);
    end;
end;

end.
