// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit osthread;

interface

uses
  {$IFDEF WIN32}windows{$ENDIF}{$IFDEF LINUX}Libc{$ENDIF}, SysUtils, Classes, orbtypes;

type
  TThreadHandle = {$IFDEF WIN32}THandle{$ENDIF}{$IFDEF LINUX}TThreadID{$ENDIF};
  TMutexAttribute = (maNormal, maRecursive);
  TMutexLockResult = (NoError, InvalidPriority, MutexAlreadyLocked);

  // This class can be used to establish a critical section. Call the
  // lock method to lock the mutex, and the unlock Method to unlock the
  // mutex. If the mutex is currently locked the thread will be
  // suspended until the mutex becomes unlocked
  TMutex = class
  private
    FCrit: TRTLCriticalSection;
  public
    constructor Create(ALocked: boolean = false);
    destructor Destroy; override;
    function trylock(): TMutexLockResult;
    procedure lock();
    procedure unlock();
  end;

  // Unlike the TMutex class this mutex can be locked recursively.
  // That is it can be locked by a thread that already has the mutex
  // locked.
  TRecursiveMutex = class
  private
    FCount: integer;
    FOwner: TThreadHandle;
    FInternal: TMutex;
    FCrit: TRTLCriticalSection;
    procedure lockI(count: integer);
    procedure unlockI();
    procedure lock(count: integer); overload;
  public
    constructor Create();
    destructor Destroy; override;
    procedure lock(); overload;
    procedure unlock();
  end;

  // This class acts like a pthreads condition variable. The difference
  // is that this implementation works with the recursive mutex
  // implementation in this library.
  TCondVar = class
  private
    FMutex: TRecursiveMutex; // Associated mutex.
{$IFDEF WIN32}
    FWaiters: integer;
    FSema: THandle; // Handle to WIN32 sempahore object.
    FBroadcastEvent: THandle;
    FHaveBroadcast: boolean;
    FWaitersMutex: TMutex;
{$ENDIF}
{$IFDEF LINUX}
    FCond: pthread_cond_t;
{$ENDIF}
    function InternalWait(tmout: Longint): boolean;
  public
    constructor Create(AMutex: TRecursiveMutex);
    destructor Destroy; override;
    function wait(): boolean; overload;
    function wait(tmout: Longint): boolean; overload;
    procedure broadcast();
    procedure signal();
  end;

  TSemaphore = class
{$IFDEF WIN32}
    FSema: THandle; // Handle to WIN32 sempahore object.
{$ENDIF}
{$IFDEF LINUX}
    //FSema: sem_t;
    FMutex: TRecursiveMutex;
    FCond: TCondVar;
    FValue: integer;
{$ENDIF}
  public
    constructor Create(val: integer = 0);
    destructor Destroy; override;
    function wait(timeout: integer = -1): boolean;//LongWord;
    procedure post();
  end;

{$DEFINE _THR_CREATE_AND_BLOCK}
  TDetachFlag = (dfDetached, dfNotDetached);
  TThreadErrorType = (teNoError, teInvalidPriority, teMutexAlreadyLocked, teNotStarted, teUnknownError);

  TThreadKey = {$IFDEF WIN32}DWORD{$ENDIF}{$IFDEF LINUX}TPThreadKey{$ENDIF};

  IORBThread = interface
  ['{7BF4A2F2-E458-4A6A-8217-888EA40B33C9}']
    procedure _run(arg: pointer);
    function Arg: Pointer;
    function ReturnValue: integer;
    procedure Finish;
    procedure wait;
    function start(arg: pointer = nil): TThreadErrorType;
    procedure _thr_startup(arg: pointer);
  end;

  TThreadSpecificCleanup = procedure(css: Pointer);

  PTKeyCleanup = ^TKeyCleanup;
  TKeyCleanup = record
    key: TThreadKey;
    release: TThreadSpecificCleanup;
    next: PTKeyCleanup;
  end;

  TThreadSpecificManager = class
  private
    FList: TThreadList;
    FMutex: TMutex;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Allocate(key: TThreadKey; clean_func: TThreadSpecificCleanup);
    procedure Release(key: TThreadKey);
    procedure Cleanup;
    class function Instance: TThreadSpecificManager;
  end;

  // This class encapsulates the functionality of a thread.
  TORBThread = class(TInterfacedObject, IORBThread)
  private
{$IFDEF WIN32}
    FThreadHandle: TThreadHandle;
    FID: THandle;
{$ENDIF}
{$IFDEF LINUX}
    FID: Cardinal;
{$ENDIF}
    FDetached: TDetachFlag;
    FReturnValue: integer;
    FNo: integer;
    FFinished: boolean;
{$IFDEF _THR_CREATE_AND_BLOCK}
    FReady: TMutex;
{$ENDIF _THR_CREATE_AND_BLOCK}
    function GetThreadId: {$IFDEF WIN32}THandle{$ENDIF}{$IFDEF LINUX}Cardinal{$ENDIF};
  protected
    FArg: pointer;
    function GetNo: integer;
    procedure SetNo(value: integer);
    { IORBThread }
    // Override this method to provide functionality to your thread.
    procedure _run(arg: pointer); virtual; abstract;
    function Arg: Pointer;
    function ReturnValue: integer;
    procedure Finish;
  public
    constructor Create(detached: TDetachFlag = dfNotDetached);
    destructor Destroy; override;
    procedure _thr_startup(arg: pointer);
    function start(arg: pointer = nil): TThreadErrorType;
    function create_thread(): TThreadErrorType;
    procedure yield();
    procedure terminate(exitval: pointer);
    procedure priority(new_priority: integer);
    procedure wait();
    // key handling
    class procedure createKey(var key: TThreadKey; clean_func: TThreadSpecificCleanup = nil);
    class procedure deleteKey(key: TThreadKey);
    class function getSpecific(key: TThreadKey): Pointer;
    class procedure setSpecific(key: TThreadKey; value: Pointer);
    //
    class function CurrentThread(): TThreadHandle; 
    property ID: {$IFDEF WIN32}THandle{$ENDIF}{$IFDEF LINUX}Cardinal{$ENDIF} read GetThreadId;
  end;

  // A monitor is used to ensure that only one thread can have access to
  // the internals of an object at one time. Use the class
  // TSynchronized to automatically lock and unlock a monitor.
  //
  // The TThreadMonitor class also provides an efficient way of notification
  // and suspension of threads. The wait method waits for a notification
  // by another thread. A thread may notify a waiting object with the
  // notify call. A thread may notify all waiting objects with the
  // notifyAll call.
  TThreadMonitor = class
  private
    FMutex: TRecursiveMutex;
    FCondVar: TCondVar;
    procedure lock();
    procedure unlock();
  public
    constructor Create;
    destructor Destroy; override;
    // Wait for notification, or timeout milliseconds.
    procedure wait(timeout: long = -1);
    // Wake one waiting object.
    procedure notify();
    // Wake all waiting objects.
    procedure notifyAll();
  end;

  ISynchronized = interface
  ['{D7BAAEFB-9F2E-4EDF-996C-43184F696CF3}']
  end;

  TLockType = (ltMonitor, ltMutex, ltRecursiveMutex);

  TSynchronized = class(TInterfacedObject, ISynchronized)
  private
    FMonitor: TThreadMonitor;
    FMutex: TMutex;
    FRecursiveMutex: TRecursiveMutex;
    FLockType: TLockType;
  public
    constructor Create(AMonitor: TThreadMonitor); overload;
    constructor Create(AMutex: TMutex); overload;
    constructor Create(AMutex: TRecursiveMutex); overload;
    destructor Destroy; override;
  end;

  TExecState = (esInit, esActive, esInitShutdown, esShutdown, esTerminated);

  TStateRefCount = class
  private
    FExecState: TExecState;
    FActiveRefs: _ulong;
    FMutex: TRecursiveMutex;
    FCondVar: TCondVar;
  public
    constructor Create;
    destructor Destroy; override;
    function activeRef(): boolean;
    procedure activeDeref();
    function StateChange(es: TExecState): boolean;
    function State(): TExecState;
    function initShutdown(): boolean;
  end;

  TRWLock = class
  private
    FLocker: TMultiReadExclusiveWriteSynchronizer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure rdlock();
    procedure wrlock();
    procedure rdunlock();
    procedure wrunlock();
  end;

  TAuto_RDLock = class(TInterfacedObject)
  private
    FRWLock: TRWLock;
  public
    constructor Create(ARWLock: TRWLock);
    destructor Destroy; override;
  end;

  TAuto_WRLock = class(TInterfacedObject)
  private
    FRWLock: TRWLock;
  public
    constructor Create(ARWLock: TRWLock);
    destructor Destroy; override;
  end;

implementation

var
  threadSpecificManager: TThreadSpecificManager;
threadvar
  threadSpecific: pointer;

{ TMutex }

constructor TMutex.Create(ALocked: boolean);
{$IFDEF LINUX}
var
  attr: Libc.TMutexAttribute;
{$ENDIF}
begin
{$IFDEF WIN32}
  InitializeCriticalSection(FCrit);
{$ENDIF}
{$IFDEF LINUX}
  pthread_mutexattr_settype(attr, PTHREAD_MUTEX_NORMAL);
  pthread_mutex_init(FCrit, attr); // not recursive
{$ENDIF}
  if ALocked then EnterCriticalSection(FCrit);
end;

destructor TMutex.Destroy;
begin
  DeleteCriticalSection(FCrit);
  inherited;
end;

// Lock the mutex.
procedure TMutex.lock;
begin
  EnterCriticalSection(FCrit);
end;

// Try lock the mutex.
function TMutex.trylock: TMutexLockResult;
begin
  if TryEnterCriticalSection(FCrit) then
    result := NoError
  else
    result := MutexAlreadyLocked;
end;

// Unlock the mutex.
procedure TMutex.unlock;
begin
  LeaveCriticalSection(FCrit);
end;

{ TCondVar }

//
// Wake all waiting threads.
//
procedure TCondVar.broadcast;
{$IFDEF WIN32}
var
  haveW: boolean;
  numW: integer;
{$ENDIF}
begin
{$IFDEF WIN32}
  FWaitersMutex.lock;
  try
    haveW := FWaiters > 0;
    numW := FWaiters;
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  if haveW then begin
    // This is a broadcast event, so set the flag and reset
    // the Event synchronization object.
    FHaveBroadcast := true;
    ResetEvent(FBroadcastEvent);
    // Release the semaphore waiters_ times.
    ReleaseSemaphore(FSema, numW, nil);
    // Wait for each thread to wake.
    // After reset the haveBroadcast_ flag.
    WaitForSingleObject(FBroadcastEvent, INFINITE);
    FHaveBroadcast := false;
  end;
{$ENDIF}
{$IFDEF LINUX}
  pthread_cond_broadcast(FCond);
{$ENDIF}
end;

constructor TCondVar.Create(AMutex: TRecursiveMutex);
begin
  FMutex := AMutex;
{$IFDEF WIN32}
  FWaiters := 0;
  FHaveBroadcast := false;;
  FSema := CreateSemaphore(nil, 0, $7fffffff, nil);
  FBroadcastEvent := CreateEvent(nil, true, false, nil);
  Assert((FSema <> INVALID_HANDLE_VALUE) or (FBroadcastEvent <> INVALID_HANDLE_VALUE));
  FWaitersMutex := TMutex.Create;
{$ENDIF}
{$IFDEF LINUX}
  pthread_cond_init(FCond, nil);
{$ENDIF}
end;

destructor TCondVar.Destroy;
begin
{$IFDEF WIN32}
  FWaitersMutex.Free;
  CloseHandle(FSema);
  CloseHandle(FBroadcastEvent);
{$ENDIF}
{$IFDEF LINUX}
  pthread_cond_destroy(FCond);
{$ENDIF}
  inherited;
end;

{$IFDEF LINUX}
type
  TCondHelper = class
  private
    FMutex: TRecursiveMutex;
    FCount: integer;
  public
    constructor Create(AMutex: TRecursiveMutex; ACount: integer);
    destructor Destroy; override;
  end;

{ TCondHelper }

constructor TCondHelper.Create(AMutex: TRecursiveMutex; ACount: integer);
begin
  FMutex := AMutex;
  FCount := ACount;
end;

destructor TCondHelper.Destroy;
begin
  // We need to acquire the mutex rationally... We unlock the
  // mutexes critical section, and lock the mutex.  Restore
  // saved information. NOTE: It's not correct to set the count
  // and the owner of the mutex, the lock API method MUST be
  // called.
  pthread_mutex_unlock(FMutex.FCrit);
  FMutex.lock(FCount);
  inherited;
end;
{$ENDIF}

function TCondVar.InternalWait(tmout: Longint): boolean;
{$IFDEF LINUX}
const
  oneBillion = 1000000000;
{$ENDIF}
var
  count: integer;
{$IFDEF WIN32}
  tm: DWORD;
  res: integer;
{$ENDIF}
{$IFDEF LINUX}
  abstime: TTimeSpec;
  tv: TTimeVal;
  unlock: TCondHelper;
  res: integer;
  tz: TTimeZone;
{$ENDIF}
begin
  // The mutex is locked at this point. We need to acquire internal,
  // save count. The cond_wait will unlock the mutexes critical
  // section.
  //
  // To see that this code is correct, consider the following cases:
  //
  // - If no thread is currently attempting to lock the mutex then code
  //   code is correct.
  //
  // - If another thread is attempting to unlock the mutex the
  //   application is in error, since the thread calling wait owns the
  //   mutex.
  //
  // - If thread B is attempting to lock the mutex, while thread A
  //   is calling wait then the following cases need to be considered:
  //
  //   - Thread B obtains lock on internal first. The mutex count
  //     will be 0, the owner will be nullThreadId. (set below)
  //     Thread B unlocks internal. Thread A can now obtain it's lock
  //     on internal. It locks internal, discovers count is 0,
  //     increments count, sets owner, and attempts to lock
  //     crit_. This lock cannot proceed until pthread_cond_wait below
  //     is called.
  //
  //   - Thread A obtains lock on internal first. Count is non-zero,
  //     and thread A does not own the mutex. Thread A attempts to
  //     lock crit_, which blocks until pthread_cond_wait below is
  //     called.
{$IFDEF LINUX}
  // Save internal state information to be restored on method
  // exit.
  FMutex.FInternal.lock();
  try
    count := FMutex.FCount;
    FMutex.FCount := 0;
    FMutex.FOwner := TORBThread.CurrentThread;
  finally
    FMutex.FInternal.unlock();
  end; { try/finally }
  // Calculate the timeout period.
  if tmout >= 0 then begin
    tz.tz_minuteswest := 0;
    tz.tz_dsttime := 0;
    gettimeofday(tv, nil);
    abstime.tv_sec := tv.tv_sec + (tmout div 1000);
    abstime.tv_nsec := (tv.tv_usec * 1000) + ((tmout mod 1000) * 1000000);
    if abstime.tv_nsec > oneBillion then begin
      Inc(abstime.tv_sec);
      abstime.tv_nsec := abstime.tv_nsec - oneBillion;
    end;
  end;
  res := 0;
  try
    unlock := TCondHelper.Create(FMutex, count);
    try
      if tmout < 0 then
        res := pthread_cond_wait(FCond, FMutex.FCrit)
      else
        res := pthread_cond_timedwait(FCond, FMutex.FCrit, abstime);
      if res <> 0 then
        raise Exception.Create('System exception');//do not localize
    finally
      unlock.Free;
    end;
    result := true;
  except
    // EINTR is translated to InterruptedException.
    if res = EINTR then
      raise Exception.Create('Interrupted exception');//do not localize
    // EAGAIN means that the wait time has passed.
    if (res = EAGAIN) or ((res = ETIMEDOUT) and (tmout >= 0)) then
      result := false
    else
      raise;
  end;
{$ENDIF}
{$IFDEF WIN32}
  // mutex is locked on entry.
  FWaiters := InterlockedIncrement(FWaiters);
  if tmout < 0 then tm := INFINITE else tm := tmout;
  // Save state information.
  FMutex.FInternal.lock();
  try
    count := FMutex.FCount;
    FMutex.FCount := 0;
    FMutex.FOwner := TORBThread.CurrentThread;
    LeaveCriticalSection(FMutex.FCrit);
  finally
    FMutex.FInternal.unlock();
  end; { try/finally }
  try
    res := WaitForSingleObject(FSema, tm);
    if res = WAIT_ABANDONED then
      raise Exception.Create('System exception');//do not localize
    result := res <> WAIT_TIMEOUT;
  except
    // Restore the mutex lock count times.
    FWaiters := InterlockedDecrement(FWaiters);
    FMutex.lock(count);
    raise;
  end; { try/except }
  // Restore the mutex lock count times.
  FWaitersMutex.lock;
  try
    Dec(FWaiters);
    if (FWaiters = 0) and FHaveBroadcast then SetEvent(FBroadcastEvent);
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  FMutex.lock(count);
{$ENDIF}
end;

procedure TCondVar.signal;
{$IFDEF WIN32}
var
  haveW: boolean;
{$ENDIF}
begin
{$IFDEF WIN32}
  FWaitersMutex.lock;
  try
    haveW := FWaiters > 0;
  finally
    FWaitersMutex.unlock;
  end; { try/finally }
  if haveW then ReleaseSemaphore(FSema, 1, nil);
{$ENDIF}
{$IFDEF LINUX}
  pthread_cond_signal(FCond);
{$ENDIF}
end;

function TCondVar.wait(tmout: Longint): boolean;
begin
  result := InternalWait(tmout);
end;

function TCondVar.wait(): boolean;
begin
  result := InternalWait(-1);
end;

{ TSemaphore }

constructor TSemaphore.Create(val: integer);
begin
{$IFDEF WIN32}
  FSema := CreateSemaphore(nil, val, $7fffffff, nil);
{$ENDIF}
{$IFDEF LINUX}
  //sem_init(FSema, false, val);
  FMutex := TRecursiveMutex.Create();
  FCond := TCondVar.Create(FMutex);
  FValue := val;
{$ENDIF}
end;

destructor TSemaphore.Destroy;
begin
{$IFDEF WIN32}
  CloseHandle(FSema);
{$ENDIF}
{$IFDEF LINUX}
  //sem_destroy(FSema);
  FMutex.Free;
  FCond.Free;
{$ENDIF}
  inherited;
end;

procedure TSemaphore.post;
begin
{$IFDEF WIN32}
  ReleaseSemaphore(FSema, 1, nil);
{$ENDIF}
{$IFDEF LINUX}
  //sem_post(FSema);
  FMutex.lock;
  try
    Inc(FValue);
  finally
    FMutex.unlock;
  end;
  FCond.signal;
{$ENDIF}
end;

function TSemaphore.wait(timeout: integer): boolean;//LongWord;
{$IFDEF LINUX}
const
  oneBillion = 1000000000;
{$ENDIF}
var
{$IFDEF WIN32}
  tm: LongWord;
{$ENDIF}
{$IFDEF LINUX}
  //abstime: timespec;
  //tv: TTimeVal;
  sync: ISynchronized;
{$ENDIF}
begin
{$IFDEF WIN32}
  if timeout = -1 then
    tm := INFINITE
  else
    tm := timeout;
  result := WaitForSingleObject(FSema, tm) = WAIT_OBJECT_0;
{$ENDIF}
{$IFDEF LINUX}
  {if timeout = -1 then begin
    result := EINTR;
    while result = EINTR do
      result := sem_wait(FSema);
  end
  else begin
    gettimeofday(tv, nil);
    abstime.tv_sec := tv.tv_sec + (timeout div 1000);
    abstime.tv_nsec := (tv.tv_usec * 1000) + ((timeout mod 1000) * 1000000);
    if abstime.tv_nsec > oneBillion then begin
      Inc(abstime.tv_sec);
      abstime.tv_nsec := abstime.tv_nsec - oneBillion;
    end;
    result := sem_timedwait(FSema, abstime);
  end;}
  result := true;
  sync := TSynchronized.Create(FMutex);
  while (FValue = 0) and result do
    result := FCond.wait(timeout);
  if result then
    Dec(FValue);
{$ENDIF}
end;

{ TORBThread }

{$IFDEF WIN32}
const
  levelMap: array [0..2] of integer = (
    THREAD_PRIORITY_LOWEST,
    THREAD_PRIORITY_NORMAL,
    THREAD_PRIORITY_HIGHEST
  );
{$ENDIF}

// Start the thread.  
procedure TORBThread._thr_startup(arg: pointer);
begin
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.lock();
{$ENDIF _THR_CREATE_AND_BLOCK}
  _run(arg);
end;

constructor TORBThread.Create(detached: TDetachFlag);
begin
  FDetached := detached;
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady := TMutex.Create;
  create_thread();
{$ENDIF _THR_CREATE_AND_BLOCK}
end;

function ThreadFunc(thr: IORBThread): integer;
var
  manager: TThreadSpecificManager;
begin
  try
    thr._thr_startup(thr.Arg);
  finally
    result := thr.ReturnValue;
    {WriteLn(IntToStr(thr._AddRef));
    thr._Release;}
    thr.Finish;
    manager := TThreadSpecificManager.Instance;
    if manager <> nil then
      manager.Cleanup();
    thr := nil;
    //EndThread(result);
  end; { try/finally }
end;

function TORBThread.create_thread: TThreadErrorType;
{$IFDEF LINUX}
var
  res: integer;
{$ENDIF}
begin
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.lock();
{$ENDIF}
{$IFDEF WIN32}
  FThreadHandle := BeginThread(nil, 0, @ThreadFunc, Pointer(Self as IORBThread), 0, FID);
  if FThreadHandle <> 0 then
    result := teNoError
  else
    result := teUnknownError;
{$ENDIF}
{$IFDEF LINUX}
  res := BeginThread(nil, @ThreadFunc, Pointer(Self as IORBThread), FID);
  if res = 0 then begin
    result := teNoError;
    if FDetached = dfDetached then
      pthread_detach(FID);
  end
  else
    result := teUnknownError;
{$ENDIF}
end;

destructor TORBThread.Destroy;
begin
  if not FFinished then
    wait();
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.unlock();
  FReady.Free;
{$ENDIF _THR_CREATE_AND_BLOCK}
  inherited;
end;

// Set the priority of this thread.
procedure TORBThread.priority(new_priority: integer);
begin
{$IFDEF WIN32}
  SetThreadPriority(FThreadHandle, levelMap[new_priority]);
{$ENDIF}
end;

// Start the thread ala JAVA:
function TORBThread.start(arg: pointer): TThreadErrorType;
begin
  FArg := arg;
{$IFDEF _THR_CREATE_AND_BLOCK}
  FReady.unlock();
  result := teNoError;
{$ELSE}
  result := create_thread();
{$ENDIF _THR_CREATE_AND_BLOCK}
end;

// Tell this thread to die.
procedure TORBThread.terminate(exitval: pointer);
begin
{$IFDEF WIN32}
  TerminateThread(FThreadHandle, PInteger(exitval)^);
{$ENDIF}
{$IFDEF LINUX}
  pthread_cancel(FID);
{$ENDIF}
end;

// Wait for this thread to finish.
procedure TORBThread.wait;
begin
{$IFDEF WIN32}
  WaitForSingleObject(FThreadHandle, INFINITE);
{$ENDIF}
{$IFDEF LINUX}
  pthread_join(FID, nil);
{$ENDIF}
end;

// Yield the current threads timeslice.
procedure TORBThread.yield;
begin
{$IFDEF WIN32}
  //yield(); //win16
  Sleep(0);
{$ENDIF}
{$IFDEF LINUX}
  sched_yield();
{$ENDIF}
end;

function TORBThread.GetThreadId: {$IFDEF WIN32}THandle{$ENDIF}{$IFDEF LINUX}Cardinal{$ENDIF};
begin
  result := FID;
end;

function TORBThread.GetNo: integer;
begin
  result := FNo;
end;

procedure TORBThread.SetNo(value: integer);
begin
  FNo := value;
end;

class procedure TORBThread.createKey(var key: TThreadKey; clean_func: TThreadSpecificCleanup);
begin
{$IFDEF WIN32}
  key := TlsAlloc();
{$ENDIF}
{$IFDEF LINUX}
  pthread_key_create(key, nil);
{$ENDIF}
  TThreadSpecificManager.Instance.Allocate(key, clean_func);
end;

class procedure TORBThread.deleteKey(key: TThreadKey);
var
  manager: TThreadSpecificManager;
begin
  // It's possible that release is called after the Thread Specific Manager
  // has already been destroyed -- although this is technically
  // an error, just ignore it.
  manager := TThreadSpecificManager.Instance;
  if manager <> nil then
    manager.Release(key);
{$IFDEF WIN32}
  TlsFree(key);
{$ENDIF}
{$IFDEF LINUX}
  pthread_key_delete(key);
{$ENDIF}
end;

class function TORBThread.getSpecific(key: TThreadKey): Pointer;
begin
{$IFDEF WIN32}
  result := TlsGetValue(key);
{$ENDIF}
{$IFDEF LINUX}
  result := pthread_getspecific(key);
{$ENDIF}
end;

class procedure TORBThread.setSpecific(key: TThreadKey; value: Pointer);
begin
{$IFDEF WIN32}
  TlsSetValue(key, value);
{$ENDIF}
{$IFDEF LINUX}
  pthread_setspecific(key, value);
{$ENDIF}
end;

function TORBThread.Arg: Pointer;
begin
  result := FArg;
end;

function TORBThread.ReturnValue: integer;
begin
  result := FReturnValue;
end;

procedure TORBThread.Finish;
begin
  FFinished := true;
end;

class function TORBThread.CurrentThread: TThreadHandle;
begin
  result := GetCurrentThreadID();
end;

{ TRWLock }

constructor TRWLock.Create;
begin
  FLocker := TMultiReadExclusiveWriteSynchronizer.Create;
end;

destructor TRWLock.Destroy;
begin
  FLocker.Free;
  inherited;
end;

procedure TRWLock.rdlock;
begin
  FLocker.BeginRead;
end;

procedure TRWLock.rdunlock;
begin
  FLocker.EndRead;
end;

procedure TRWLock.wrlock;
begin
  FLocker.BeginWrite;
end;

procedure TRWLock.wrunlock;
begin
  FLocker.EndWrite;
end;

{ TAuto_RDLock }

constructor TAuto_RDLock.Create(ARWLock: TRWLock);
begin
  FRWLock := ARWLock;
  FRWLock.rdlock();
end;

destructor TAuto_RDLock.Destroy;
begin
  FRWLock.rdunlock();
  inherited;
end;

{ TAuto_WRLock }

constructor TAuto_WRLock.Create(ARWLock: TRWLock);
begin
  FRWLock := ARWLock;
  FRWLock.wrlock();
end;

destructor TAuto_WRLock.Destroy;
begin
  FRWLock.wrunlock();
  inherited;
end;

{ TStateRefCount }

procedure TStateRefCount.activeDeref;
begin
  FMutex.lock();
  try
    FActiveRefs := Pred(FActiveRefs);
    if (FActiveRefs = 0) and (FExecState <> esActive) then
      FCondVar.signal();
  finally
    FMutex.unlock();
  end; { try/finally }
end;

function TStateRefCount.activeRef: boolean;
begin
  FMutex.lock();
  try
    result := FExecState = esActive;
    if result then FActiveRefs := Succ(FActiveRefs);
  finally
    FMutex.unlock();
  end; { try/finally }
end;

constructor TStateRefCount.Create;
begin
  FMutex := TRecursiveMutex.Create;
  FCondVar := TCondVar.Create(FMutex);
  FExecState := esInit;
  FActiveRefs := 0;
end;

destructor TStateRefCount.Destroy;
begin
  FCondVar.Free;
  FMutex.Free;
  inherited;
end;

function TStateRefCount.initShutdown: boolean;
begin
  FMutex.lock();
  try
    result := FExecState = esActive;
    if result then FExecState := esInitShutdown;
  finally
    FMutex.unlock();
  end; { try/finally }
end;

function TStateRefCount.State: TExecState;
begin
  result := FExecState;
end;

function TStateRefCount.StateChange(es: TExecState): boolean;
begin
  FMutex.lock();
  try
    result := es > FExecState;
    if result then begin
      FExecState := es;
      // wait until the state change has become effective
      while FActiveRefs <> 0 do
        FCondVar.wait();//FMutex);
    end;
  finally
    FMutex.unlock();
  end; { try/finally }
end;

{ TThreadSpecificManager }

procedure TThreadSpecificManager.Allocate(key: TThreadKey;
  clean_func: TThreadSpecificCleanup);
var
  node: PTKeyCleanup;
  list: TList;
  sync: ISynchronized;
begin
  sync := TSynchronized.Create(FMutex);
  list := FList.LockList;
  try
    node := AllocMem(SizeOf(TKeyCleanup));
    node^.key := key;
    node^.release := clean_func;
    list.Add(node);
  finally
    FList.UnlockList;
  end; { try/finally }
  {node := AllocMem(SizeOf(TKeyCleanup));
  node^.key := key;
  node^.release := clean_func;
  node^.next := FHead;
  FHead := node;}
end;

procedure TThreadSpecificManager.Cleanup;
var
  list: TList;
  i: integer;
  data: Pointer;
  node: PTKeyCleanup;
  cleanfunc: TThreadSpecificCleanup;
begin
  list := FList.LockList;
  try
    i := list.Count - 1;
    while i >= 0 do begin
      node := PTKeyCleanup(list.Items[i]);
      data := TORBThread.getSpecific(node^.key);
      cleanfunc := node^.release;
      cleanfunc(data);
      TORBThread.setSpecific(node^.key, nil);
      i := Pred(i);
    end; { while }
  finally
    FList.UnlockList;
  end; { try/finally }
end;

constructor TThreadSpecificManager.Create;
begin
  FMutex := TMutex.Create;
  FList := TThreadList.Create;
end;

destructor TThreadSpecificManager.Destroy;
begin
  FMutex.Free;
  FList.Free;
  inherited;
end;

class function TThreadSpecificManager.Instance: TThreadSpecificManager;
begin
  if threadSpecificManager = nil then
    threadSpecificManager := TThreadSpecificManager.Create;
  result := threadSpecificManager;
end;

procedure TThreadSpecificManager.Release(key: TThreadKey);
var
  list: TList;
  i: integer;
begin
  list := FList.LockList;
  try
    i := list.Count - 1;
    while i >= 0 do begin
      if PTKeyCleanup(list.Items[i])^.key = key then
        Break;
      i := Pred(i);
    end; { while }
    if i >= 0 then begin
      FreeMem(list.Items[i], SizeOf(TKeyCleanup));
      list.Delete(i);
    end;
  finally
    FList.UnlockList;
  end; { try/finally }
end;

{ TThreadMonitor }

constructor TThreadMonitor.Create;
begin
  FMutex := TRecursiveMutex.Create();
  FCondVar := TCondVar.Create(FMutex);
end;

destructor TThreadMonitor.Destroy;
begin
  FMutex.Free;
  FCondVar.Free;
  inherited;
end;

procedure TThreadMonitor.lock;
begin
  FMutex.lock();
end;

procedure TThreadMonitor.notify;
begin
  {FMutex.lock();
  try}
    FCondVar.signal();
  {finally
    FMutex.unlock();
  end; { try/finally }
end;

procedure TThreadMonitor.notifyAll;
begin
  {FMutex.lock();
  try}
    FCondVar.broadcast();
  {finally
    FMutex.unlock();
  end; { try/finally }
end;

procedure TThreadMonitor.unlock;
begin
  FMutex.unlock();
end;

procedure TThreadMonitor.wait(timeout: long);
begin
  {FMutex.lock();
  try}
    FCondVar.wait(timeout);
  {finally
    FMutex.unlock();
  end; { try/finally }
end;

{ TSynchronized }

constructor TSynchronized.Create(AMonitor: TThreadMonitor);
begin
  FLockType := ltMonitor;
  FMonitor := AMonitor;
  FMonitor.lock();
end;

constructor TSynchronized.Create(AMutex: TMutex);
begin
  FLockType := ltMutex;
  FMutex := AMutex;
  FMutex.lock();
end;

constructor TSynchronized.Create(AMutex: TRecursiveMutex);
begin
  FLockType := ltRecursiveMutex;
  FRecursiveMutex := AMutex;
  FRecursiveMutex.lock();
end;

destructor TSynchronized.Destroy;
begin
  case FLockType of
    ltMonitor: FMonitor.unlock();
    ltMutex: FMutex.unlock();
    ltRecursiveMutex: FRecursiveMutex.unlock();
  end;
  inherited;
end;

{ TRecursiveMutex }

constructor TRecursiveMutex.Create;
{$IFDEF LINUX}
var
  attr: Libc.TMutexAttribute;
{$ENDIF}
begin
  FInternal := TMutex.Create;
{$IFDEF WIN32}
  InitializeCriticalSection(FCrit);
{$ENDIF}
{$IFDEF LINUX}
  //attr.__mutexkind := 0;
  pthread_mutexattr_settype(attr, PTHREAD_MUTEX_RECURSIVE{PTHREAD_MUTEX_NORMAL});
  pthread_mutex_init(FCrit, attr); // not recursive
{$ENDIF}
end;

destructor TRecursiveMutex.Destroy;
begin
  FInternal.Free;
  DeleteCriticalSection(FCrit);
  inherited;
end;

procedure TRecursiveMutex.lock;
begin
  // Work around lack of mutable.
  Self.lockI(1);
end;

procedure TRecursiveMutex.lock(count: integer);
begin
  // Work around lack of mutable.
  Self.lockI(count);
end;

procedure TRecursiveMutex.lockI(count: integer);
var
  ob: boolean;
begin
  // This flag is set to true once the mutex has been obtained.
  ob := false;
  while not ob do begin
    // Lock the internal mutex.
    FInternal.lock();
    // count_ represents the number of times the mutex has been
    // aquired.  If count_ is zero then the mutex is not yet
    // aquired.  Note that two mutexes cannot be in here at once,
    // since this CS is protected by the mutex internal_.
    if FCount = 0 then begin
      // We're attempting to aquire the mutex count more times.
      FCount := count;
      FOwner := TORBThread.CurrentThread();
      ob := true;
      // Acquire the mutexes critical section.
      try
        EnterCriticalSection(FCrit);
      except
        try
          FInternal.unlock();
        except
        end; { try/except }
        raise;
      end; { try/except }
    end
    // Otherwise count_ is non-zero.  This means that the mutex is
    // currently aquired.  There are two cases here:
    //
    // - The mutex is owned by the caller. Simply increment the
    //   count in this case.
    //
    // - The mutex is owned by a thread other than the caller. In
    //   this case we attempt to aquire the mutexes critical
    //   section. This CS is unlocked once mutex is released count_
    //   times.
    else if FOwner = TORBThread.CurrentThread() then begin
      FCount := Succ(FCount);
      ob := true;
    end;
    FInternal.unlock();
    // If we haven't yet obtained the mutex lock the critical
    // section mutex.  Immediately unlock the CS, then attempt to
    // re-obtain.
    if not ob then begin
      EnterCriticalSection(FCrit);
      LeaveCriticalSection(FCrit);
    end;
  end; { while }
end;

procedure TRecursiveMutex.unlock;
begin
  // Work around lack of mutable.
  Self.unlockI();
end;

procedure TRecursiveMutex.unlockI;
begin
  // Aquire the internal mutex.
  FInternal.lock();
  // If count_ decrements to zero then unlock the mutexes CS. The
  // mutex is no longer owned by a thread.
  FCount := Pred(FCount);
  if FCount = 0 then begin
    FOwner := TORBThread.CurrentThread();
    LeaveCriticalSection(FCrit);
  end;
  FInternal.unlock();
end;

initialization
finalization
  FreeAndNil(threadSpecificManager);
end.

